// ############################################################################################
//  VirtualWalk2.cc
//  A demo application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

//#define __HIGH_RESOLUTION__

// Common headers for demo application
#include "DemoCommon.h"


// Just for shortcut
typedef CBasicApplication<> MyApplicationBase;

// Specialized basic application class
class MyApplication : public MyApplicationBase {
  public:
    MyApplication();
    ~MyApplication() throw();
    virtual void EventKeyPress(EKeySymbol eKeySym);
    virtual void EventKeyRelease(EKeySymbol eKeySym);
    virtual void EventUpdateFrame(bool bWindowIsActive);
    virtual void EventCloseWindow();

    inline bool detectCollision(const float afCurPosXYZ[3], float afTransFactorXYZ[3]) {
      if(csTreeSpr.Collide(afCurPosXYZ, afTransFactorXYZ)) return(false);
      if(csGround.PlaneCollide(afCurPosXYZ[0] + afTransFactorXYZ[0],
                               afCurPosXYZ[1] + afTransFactorXYZ[1],
                               afCurPosXYZ[2] + afTransFactorXYZ[2],
                               5.0f) != ECT_None ) return(false);
      return(true);
    }

  private:
    static const int32_t NormalViewHeight = 65;

    CPanel3DBuffer        csGround, csWater;
    CWalkingCameraControl csCamera;
    CImageBuilder         csImgBuild;
    CImage                csImgTree, csImgSceneryN, csImgSceneryB;
    CAnimatedImage        csImgLight;
    CTexture              csTGround;
    CAnimatedTexture      csTWater;
    CSprite2D             csTreeSpr, csLightSpr;
    CSnow2D               csSnow;
    CSoundBuffer         *pcsSBWalk;
    CSoundBuffer         *pcsSBJump;
    CSoundBuffer         *pcsSBWater;
    int32_t               n32Skip, n32SkipMax;              // For frame skipping
    bool                  bForward, bBackward;              // For animation
    bool                  bLeft, bRight, bSL, bSR;          // ---
    bool                  bRun, bLU, bLD, bBio;             // ---
    bool                  bUseSceneryImage;                 // Shall we use image for scenery?
    int32_t               n32Jump, n32C;                    // ---
    float                 fLX[4], fLZ[4], fLDX[4], fLDZ[4]; // ---
    CPoint3D              csSnowXSpeed;                     // Snow particle X speed
};


// Callback functions
static bool WCCCallback(EWCCContext eContext, CWalkingCameraControl *pcsWCC, const float afCurPosXYZ[3], float afTransFactorXYZ[3], void *pvUserData)
{
  UNUSED_PARAMS_2(eContext, pcsWCC);
  MyApplication *pcsMyApp = reinterpret_cast<MyApplication*>(pvUserData);
  return(pcsMyApp->detectCollision(afCurPosXYZ, afTransFactorXYZ));
}


// Main program
int main()
{
  MyApplication csMyApp;
  return(csMyApp.DoMainLoop());
}


// MyApplication implementation
MyApplication::MyApplication()
#ifdef __HIGH_RESOLUTION__
: MyApplicationBase("Demo : Virtual Walk 2", NULL, 640, 480, 640, 480),
#else
: MyApplicationBase("Demo : Virtual Walk 2", NULL, 640, 480, 320, 240),
#endif
  pcsSBWalk(NULL), pcsSBJump(NULL), pcsSBWater(NULL),
  n32Skip(0), n32SkipMax(0), bUseSceneryImage(true), csSnowXSpeed(2.0f, 0.0f, 0.0f)
{
  // Tell the user to be patient
  ShowInitSystemPleaseWaitMessage(iVideoEnv());

  // Set wanted FPS
  SetWantedFPS(50.0f);

  // Load audios
  if(!AudioIsDisabled()) {
    pcsSBWalk = iAudioEnv().NewSoundBuffer();
    pcsSBWalk->LoadWAV("../media/wav/walk_w.wav");
    pcsSBWalk->AdjustAmplitude(0.6f, 0.6f);
    pcsSBJump = iAudioEnv().NewSoundBuffer();
    pcsSBJump->LoadWAV("../media/wav/jump.wav");
    pcsSBJump->AdjustAmplitude(0.5f, 0.5f);
    pcsSBWater = iAudioEnv().NewSoundBuffer();
    pcsSBWater->LoadWAV("../media/wav/water.wav");
    pcsSBWater->AdjustAmplitude(0.2f, 0.2f);
  }
  if(!MidiIsDisabled()) iMidiPlayer().LoadMID("../media/midi/ame002.mid");

  // Load ground texture
  try {
    csImgBuild.LoadJPG("../media/image/tex2.jpg", true);
    //csImgBuild.LoadPNG("/home/aloy/Images/konsole-vge.png", true);
    csImgBuild.Resize(256, 256);
    //csImgBuild.Resize(512, 512);
    csImgBuild.SmearEffect(2);
  }
  catch(...) {
    csImgBuild.CreateNoiseImage(256, 256, 16, 18, 16, 5, 5, 2);
    csImgBuild.OilPaintingEffect(4);
    csImgBuild.BlurEffect(1);
  }
  csImgBuild.TransferToTexture(csTGround);

  // Load / create water texture
  try {
    csImgBuild.LoadVIB("../.mediaCaches/water.vib");
  }
  catch(...) {
    csImgBuild.CreateNoiseImage(160, 140, 8, 8, 16, 3, 3, 3);
    csImgBuild.Resize(1280, 1280);
    csImgBuild.SmearEffect();
    csImgBuild.SaveVIB("../.mediaCaches/water.vib");
  }
  csImgBuild.TransferToAnimatedTexture(csTWater, 5, 5);

  // Load tree sprite image
  csImgBuild.LoadBMP("../media/image/tree1.bmp", true);
  csImgBuild.RoughResize(500, 800);
  csImgBuild.TransferToImage(csImgTree);

  // Load the light sprite image
  csImgBuild.LoadBMP("../media/image/sparklight.bmp", true);
  csImgBuild.AdjustBrightness(0.5, 0, true, true, false);
  csImgBuild.Resize(320*2, 32*2);
  csImgBuild.TransferToAnimatedImage(csImgLight, 1, 10);

  // Load the skybox image #1
  csImgBuild.LoadBMP("../media/image/sky.bmp", true);
  csImgBuild.ConvertRGB2YUV();
  csImgBuild.YUVScaleUV(0.25f);
  csImgBuild.YUVScaleY(0.55f);
  csImgBuild.ConvertYUV2RGB();
  csImgBuild.Resize(i3DEnv().RenderBufferWidth()*2, i3DEnv().RenderBufferHeight()/4, true);
  csImgBuild.TransferToImage(csImgSceneryN);

  // Create the skybox image #2
  csImgBuild.AdjustBrightness(1.15f, 0, true, false, false);
  csImgBuild.AdjustBrightness(0.25f, 0, false, true, true);
  csImgBuild.TransferToImage(csImgSceneryB);

  // Free image
  csImgBuild.FreeImage();

  // Create scene
  i3DEnv().SetFogState(12, 12, 12, 12);
  i3DEnv().SetViewport(0, 0, i3DEnv().RenderBufferWidth(), i3DEnv().RenderBufferHeight());
  i3DEnv().SetZClipping(50, 1500);
  i3DEnv().SetDefaultOmniLightPosition(300, 600, 1100);
  i3DEnv().SetDefaultOmniLightColor(20, 20, 18);

  float groundVertices[57 * 3] = {
    -400,-20,1200,   -200,-20,1200,   0,-20,1200,   200,-20,1200,   400,-20,1200,
    -400,-20,1000,   -200,-20,1000,   0,-20,1000,   200,-20,1000,   400,-20,1000,
    -400,-20, 800,   -200,-20, 800,   0,-20, 800,   200,-20, 800,   400,-20, 800,
    -400,-20, 400,   -200,-20, 400,   0,-20, 400,   200,-20, 400,   400,-20, 400,
    -400,-20, 200,   -200,-20, 200,   0,-20, 200,   200,-20, 200,   400,-20, 200,
    -400,180,1200,   -200,180,1200,   0,180,1200,   200,180,1200,   400,180,1200,
    -400,180,1000,                                                  400,160,1000,
    -400,180, 800,                                                  400,140, 800,
    -400,180, 400,                                                  400,120, 400,
    -400,170, 200,   -200,180, 200,   0,180, 200,   200,180, 200,   400,100, 200,
    -550,680,1350,   -275,520,1350,   0,440,1350,   275,400,1350,   550,380,1350,
    -550,520,1075,                                                  550,370,1075,
    -550,440, 800,                                                  550,350, 800,
    -550,400, 325,                                                  550,310, 325,
    -550,380,  50,   -275,370,  50,   0,350,  50,   275,310,  50,   550,230,  50 };
  uint32_t groundPanels[48 * 4] = {
     0, 1, 6, 5,    1, 2, 7, 6,    2, 3, 8, 7,    3, 4, 9, 8,
     5, 6,11,10,    6, 7,12,11,    7, 8,13,12,    8, 9,14,13,
    10,11,16,15,   11,12,17,16,   12,13,18,17,   13,14,19,18,
    15,16,21,20,   16,17,22,21,   17,18,23,22,   18,19,24,23,
    25,26, 1, 0,   26,27, 2, 1,   27,28, 3, 2,   28,29, 4, 3,
    29,31, 9, 4,   31,33,14, 9,   33,35,19,14,   35,40,24,19,
    40,39,23,24,   39,38,22,23,   38,37,21,22,   37,36,20,21,
    36,34,15,20,   34,32,10,15,   32,30, 5,10,   30,25, 0, 5,
    41,42,26,25,   42,43,27,26,   43,44,28,27,   44,45,29,28,
    45,47,31,29,   47,49,33,31,   49,51,35,33,   51,56,40,35,
    56,55,39,40,   55,54,38,39,   54,53,37,38,   53,52,36,37,
    52,50,34,36,   50,48,32,34,   48,46,30,32,   46,41,25,30 };
  csGround.InitFromMemory(groundVertices, 57, groundPanels, 48, true, true);
  csGround.SetColor(31, 0, 0);
  csGround.SetTexture(&csTGround);
  csGround.CalcDefTexCoords();

  float waterVertices[25 * 3] = {
    -400,20,1200,   -200,20,1200,   0,20,1200,   200,20,1200,   400,20,1200,
    -400,20,1000,   -200,20,1000,   0,20,1000,   200,20,1000,   400,20,1000,
    -400,20, 800,   -200,20, 800,   0,20, 800,   200,20, 800,   400,20, 800,
    -400,20, 400,   -200,20, 400,   0,20, 400,   200,20, 400,   400,20, 400,
    -400,20, 200,   -200,20, 200,   0,20, 200,   200,20, 200,   400,20, 200 };
  uint32_t waterPanels[16 * 4] = {
     0, 1, 6, 5,    1, 2, 7, 6,    2, 3, 8, 7,    3, 4, 9, 8,
     5, 6,11,10,    6, 7,12,11,    7, 8,13,12,    8, 9,14,13,
    10,11,16,15,   11,12,17,16,   12,13,18,17,   13,14,19,18,
    15,16,21,20,   16,17,22,21,   17,18,23,22,   18,19,24,23 };
  csWater.InitFromMemory(waterVertices, 25, waterPanels, 16);
  csWater.SetColor(8, 8, 16);
  csWater.SetTexture(&csTWater);
  csWater.CalcDefTexCoords();

  float afTPos[] = { -300, -20, 320,   -100, -20, 900,   100, -20, 900,   300, -20, 320,
                        0, -20, 320,   -300, -20, 610,   300, -20, 610,     0, -20, 610 };
  csTreeSpr.Initialize(afTPos, 8, &csImgTree);
  csTreeSpr.SetColor(0, 31, 0);

  fLX[0] = -200; fLZ[0] = 300; fLDX[0] =  2; fLDZ[0] = -2;
  fLX[1] =  300; fLZ[1] = 500; fLDX[1] = -2; fLDZ[1] =  2;
  fLX[2] = -300; fLZ[2] = 700; fLDX[2] = -2; fLDZ[2] =  2;
  fLX[3] =  200; fLZ[3] = 900; fLDX[3] =  2; fLDZ[3] = -2;
  float afLPos[] = {0,0,0, 0,0,0, 0,0,0, 0,0,0};
  csLightSpr.Initialize(afLPos, 4, &csImgLight, EP_CenterCenter);

  csSnow.Initialize(&i3DEnv(), 1000, 500.0f, -1000, 1320, -500, 0, i3DEnv().RenderBufferHeight()-40, i3DEnv().RenderBufferHeight()-10, 0.0f, 2.0f, 31, 31, 31);

  // Initialize camera
  csCamera.SetCallback(WCCCallback, this);
  csCamera.SetPosition(-200, NormalViewHeight, 400);
  csCamera.SetYRotation(0);
  csCamera.UpdateCameraMatrix();
  bForward = bBackward = bLeft = bRight = bRun = bLU = bLD = bSL = bSR = bBio = false;
  n32Jump = 0;
  csCamera.iSkybox().SetImageSource(csImgSceneryN);
  csCamera.iSkybox().SetCameraYOffset(NormalViewHeight);

  // Update snow particle X speed for the first time
  csCamera.GetYCameraMatrix().Transform(csSnowXSpeed);
  csSnow.SetXSpeed(csSnowXSpeed.GetWorldX());

  // Play the MIDI file
  if(!MidiIsDisabled()) {
    iMidiPlayer().SetVolume(30);
    iMidiPlayer().Play(true);
  }

  // Play the water sound
  pcsSBWater->Play(true);
}

MyApplication::~MyApplication() throw()
{
  delete pcsSBWalk;
  delete pcsSBJump;
  delete pcsSBWater;
}

void MyApplication::EventKeyPress(EKeySymbol eKeySym)
{
       if(eKeySym == '\\')       iVideoEnv().SetFullscreen(!iVideoEnv().IsFullscreen());
  else if(eKeySym == VKS_ESC)    EventCloseWindow();
  else if(eKeySym == VKS_LEFT)   bLeft = (bRight) ? false : true;
  else if(eKeySym == VKS_RIGHT)  bRight = (bLeft) ? false : true;
  else if(eKeySym == VKS_UP)     bForward = (bBackward) ? false : true;
  else if(eKeySym == VKS_DOWN)   bBackward = (bForward) ? false : true;
  else if(eKeySym == VKS_LSHIFT) bRun = true;
  else if(eKeySym == 'a')        bLU = (bLD) ? false : true;
  else if(eKeySym == 'z')        bLD = (bLU) ? false : true;
  else if(eKeySym == 'q')        bSL = (bSR) ? false : true;
  else if(eKeySym == 'w')        bSR = (bSL) ? false : true;
  else if(eKeySym == ' ') {
    if(AudioIsDisabled() && (n32Jump <= 0))
      n32Jump = 4;
    else if(!pcsSBJump->IsPlaying() && (n32Jump <= 0)) {
        pcsSBJump->Play();
        n32Jump = 4;
    }
  }
}

void MyApplication::EventKeyRelease(EKeySymbol eKeySym)
{
       if(eKeySym == VKS_LEFT)   bLeft = false;
  else if(eKeySym == VKS_RIGHT)  bRight = false;
  else if(eKeySym == VKS_UP)     bForward = false;
  else if(eKeySym == VKS_DOWN)   bBackward = false;
  else if(eKeySym == VKS_LSHIFT) bRun = false;
  else if(eKeySym == 'a')        bLU = false;
  else if(eKeySym == 'z')        bLD = false;
  else if(eKeySym == 'q')        bSL = false;
  else if(eKeySym == 'w')        bSR = false;
  else if(eKeySym == 'b') {
    bBio = !bBio;
    if(bBio) {
      i3DEnv().SetFogState(12, 4, 4, 12);
      csCamera.iSkybox().SetImageSource(csImgSceneryB);
      csSnow.SetColor(31, 12, 12);
    }
    else {
      i3DEnv().SetFogState(12, 12, 12, 12);
      csCamera.iSkybox().SetImageSource(csImgSceneryN);
      csSnow.SetColor(31, 31, 31);
    }
  }
  else if(eKeySym == 's')
    bUseSceneryImage = !bUseSceneryImage;
  else if(eKeySym == '[') {
    n32SkipMax = (n32SkipMax > 0) ? (n32SkipMax - 1) : 0;
    if(n32SkipMax == 0) EnableFPSControl();
  }
  else if(eKeySym == ']') {
    n32SkipMax = (n32SkipMax < 9) ? (n32SkipMax + 1) : 9;
    DisableFPSControl();
  }
}

void MyApplication::EventUpdateFrame(bool bWindowIsActive)
{
  // Update frame (if the application is not active)
  if(!bWindowIsActive) {
    iVideoEnv().PutRenderImageOnDisplay();
    iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
    iVideoEnv().UseDefaultBigFont();
    iVideoEnv().DrawText(0, 150, "Inactive State", ETA_CenterScreen);
    iVideoEnv().UseDefaultSmallFont();
    iVideoEnv().UpdateDisplay();
    return;
  }

  // Update Y rotation and XY position
  if(!bRun) {
    if(bLeft)     csCamera.RotateLeft(10);
    if(bRight)    csCamera.RotateRight(10);
    if(bForward)  csCamera.StepForward(4);
    if(bBackward) csCamera.StepBackward(4);
    if(bSL)       csCamera.StrafeLeft(4);
    if(bSR)       csCamera.StrafeRight(4);
  }
  else {
    if(bLeft)     csCamera.RotateLeft(20);
    if(bRight)    csCamera.RotateRight(20);
    if(bForward)  csCamera.StepForward(8);
    if(bBackward) csCamera.StepBackward(8);
    if(bSL)       csCamera.StrafeLeft(8);
    if(bSR)       csCamera.StrafeRight(8);
  }
  // Update Y position
  if(n32Jump) {
    n32Jump--;
    csCamera.ClimbUp(10);
  }
  else if(csCamera.GetCurYPos() > NormalViewHeight) {
    csCamera.ClimbDown(5);
  }
  // Update X rotation
  if(bLU)
    csCamera.LookUp(150);
  else if(bLD)
    csCamera.LookDown(150);
  else
    csCamera.LookStraight();

  // Update sprites position
  for(n32C = 0; n32C < 4; n32C++) {
    fLX[n32C] += fLDX[n32C];
    if(fLX[n32C] < -300.0f) { fLX[n32C] = -300.0f; fLDX[n32C] = -fLDX[n32C]; }
    if(fLX[n32C] >  300.0f) { fLX[n32C] =  300.0f; fLDX[n32C] = -fLDX[n32C]; }
    fLZ[n32C] += fLDZ[n32C];
    if(fLZ[n32C] <  300.0f) { fLZ[n32C] =  300.0f; fLDZ[n32C] = -fLDZ[n32C]; }
    if(fLZ[n32C] > 1100.0f) { fLZ[n32C] = 1100.0f; fLDZ[n32C] = -fLDZ[n32C]; }
    csLightSpr.SetPosition(n32C, fLX[n32C], 65, fLZ[n32C]);
  }

  // Update textures
  csImgLight.StepFrame(5);
  csTWater.StepFrame(5);

  // Update snow particles
  csCamera.GetYCameraMatrix().Transform(csSnowXSpeed);
  csSnow.SetXSpeed(csSnowXSpeed.GetWorldX());
  csSnow.Update();

  // Render scene
  if(n32Skip < n32SkipMax)
    n32Skip++;
  else {
    // Reset counter
    n32Skip = 0;

    // Update matrix
    csCamera.UpdateCameraMatrix();

    // Clear scene with background
    if(!bUseSceneryImage) {
      if(bBio)
        i3DEnv().ClearRenderBuffer(VColor::Maroon);
      else
        i3DEnv().ClearRenderBuffer(VColor::DimGray);
    }
    else {
      csCamera.iSkybox().Render();
    }
    i3DEnv().ClearZBuffer();

    // Render the scene
    if(bBio) {
      csGround.FogTCRender(csCamera.GetCameraMatrix());
      csTreeSpr.FogSBRender(csCamera.GetCameraMatrix());
    }
    else {
      //csGround.FogTGRender(csCamera.GetCameraMatrix());
      csGround.FogTNRender(csCamera.GetCameraMatrix());
      csTreeSpr.FogSNRender(csCamera.GetCameraMatrix());
    }
    csWater.FogTNTransRender(csCamera.GetCameraMatrix());
    csLightSpr.ARender(csCamera.GetCameraMatrix());
    csSnow.FogSRenderWZ();

    //Update info
    char acBuff[1024];
    snprintf(acBuff, sizeof(acBuff), "%s, %d skips (%+08.2f, %+08.2f, %+08.2f, %+05d, %+05d) <S%+04.2f>",
             GetFPSInfoString(), n32SkipMax,
             csCamera.GetCurXPos(), csCamera.GetCurYPos(), csCamera.GetCurZPos(),
             csCamera.GetCurXRot(), csCamera.GetCurYRot(),
             csSnowXSpeed.GetWorldX());

    // Update display
    iVideoEnv().PutRenderImageOnDisplay();
    iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
    iVideoEnv().DrawText(10, 20, acBuff);
    iVideoEnv().DrawText(10,  50, "Press 'ARROW' to move around");
    iVideoEnv().DrawText(10,  65, "Press 'SHIFT' to to run");
    iVideoEnv().DrawText(10,  80, "Press 'SPACE' to jump");
    iVideoEnv().DrawText(10,  95, "Press 'A' to look up");
    iVideoEnv().DrawText(10, 110, "Press 'Z' to look down");
    iVideoEnv().DrawText(10, 125, "Press 'Q' to strafe left");
    iVideoEnv().DrawText(10, 140, "Press 'W' to strafe right");
    if(bBio)
      iVideoEnv().DrawText(10, 155, "Press 'B' to release bio-visor");
    else
      iVideoEnv().DrawText(10, 155, "Press 'B' to use bio-visor");
    if(bUseSceneryImage)
      iVideoEnv().DrawText(10, 185, "Press 'S' to disable skybox");
    else
      iVideoEnv().DrawText(10, 185, "Press 'S' to enable skybox");
    iVideoEnv().DrawText(10, 215, "Press '[' to skip less frames");
    iVideoEnv().DrawText(10, 230, "Press ']' to skip more frames");
    iVideoEnv().DrawText(10, 260, "Press escape to exit ...");
    iVideoEnv().UpdateDisplay();
  }

  // Update audio
  if(!AudioIsDisabled()) {
    if((bLeft || bRight || bForward || bBackward || bSL || bSR) &&
       (!pcsSBWalk->IsPlaying())) pcsSBWalk->Play();
    iAudioEnv().UpdateBuffer();
  }
}

void MyApplication::EventCloseWindow()
{
  iVideoEnv().MessageBox("End Program", "Click \"OK\" to exit ...");
  MyApplicationBase::EventCloseWindow();
}
