/** \file ComplexNumberBasedFractals.h */
// ############################################################################################
//  ComplexNumberBasedFractals.h
//  Definition & implementation of various CCNBFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Pemrograman Fraktal Resolusi Tinggi untuk Kartu Trident SVGA
//    Ir. P. Insap Santosa, M.Sc. & Ir. R. Tory K.
//    Penerbit ANDI Yogyakarta
//    Indonesia, 1997
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___FRACTAL___COMPLEX_NUMBER_BASED_FRACTALS_H_
#define _2D_MANIPULATION___FRACTAL___COMPLEX_NUMBER_BASED_FRACTALS_H_


// VortexGE headers
#include "DefiniteAreaFractals.h"

namespace vge {

  //!
  //! \brief Fractal that will draw Mandelbrot set
  //!
  //! \api experiment
  class VGE_EXPORT FractalMandelbrot : public CCNBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow();
      // Seed selection funtions
      virtual void SelectDefaultSeed() {};

    protected:
      // Equation iteration functions
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter);
  };


  //!
  //! \brief Fractal that will draw Julia set
  //!
  //! \api experiment
  class VGE_EXPORT FractalJulia : public CCNBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow();
      // Seed selection funtions
      virtual void SelectDefaultSeed();

    protected:
      // Equation iteration functions
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter);
  };


  //!
  //! \brief Fractal that will draw Dragon curve
  //!
  //! \api experiment
  class VGE_EXPORT FractalDragon : public CCNBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow();
      // Seed selection funtions
      virtual void SelectDefaultSeed();

    protected:
      // Equation iteration functions
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter);
  };


  //!
  //! \brief Fractal that will draw Sierspinski's triangle
  //!
  //! \api experiment
  class VGE_EXPORT FractalSierspinski : public CCNBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow();
      // Seed selection funtions
      virtual void SelectDefaultSeed() {};

    protected:
      // Equation iteration functions
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter);
  };


  //!
  //! \brief Fractal that will draw sun-like image
  //!
  //! \api experiment
  class VGE_EXPORT FractalSun : public CCNBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow();
      // Seed selection funtions
      virtual void SelectDefaultSeed() {};

    protected:
      // Equation iteration functions
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter);
  };

  //!
  //! \brief Fractal that will draw the result of complex number iteration using sine function
  //!
  //! \api experiment
  class VGE_EXPORT FractalFSin : public CTFBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow() { SelectWindow(0.00f, -5.00f, 12.60f, 5.00f); }

    protected:
      // Equation iteration functions
      virtual void CalculateEqu(complex<float> &tZ, const complex<float> &tC);
  };


  //!
  //! \brief Fractal that will draw the result of complex number iteration using cosine function
  //!
  //! \api experiment
  class VGE_EXPORT FractalFCos : public CTFBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow() { SelectWindow(-1.75f, -5.00f, 10.95f, 5.00f); }

    protected:
      // Equation iteration functions
      virtual void CalculateEqu(complex<float> &tZ, const complex<float> &tC);
  };


  //!
  //! \brief Fractal that will draw the result of complex number iteration using hyperbolic sine function
  //!
  //! \api experiment
  class VGE_EXPORT FractalFSinh : public CTFBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow() { SelectWindow(-5.00f, 0.00f, 5.00f, 12.60f); }

    protected:
      // Equation iteration functions
      virtual void CalculateEqu(complex<float> &tZ, const complex<float> &tC);
  };


  //!
  //! \brief Fractal that will draw the result of complex number iteration using hyperbolic cosine function
  //!
  //! \api experiment
  class VGE_EXPORT FractalFCosh : public CTFBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow() { SelectWindow(-6.00f, -3.00f, 4.00f, 9.40f); }

    protected:
      // Equation iteration functions
      virtual void CalculateEqu(complex<float> &tZ, const complex<float> &tC);
  };


  //!
  //! \brief Fractal that will draw the result of complex number iteration using exponential function
  //!
  //! \api experiment
  class VGE_EXPORT FractalFExp : public CTFBFractal {
    public:
      // Window selection funtions
      virtual void SelectDefaultWindow() { SelectWindow(-6.00f, -15.00f, 3.00f, 15.00f); }

    protected:
      // Equation iteration functions
      virtual void CalculateEqu(complex<float> &tZ, const complex<float> &tC);
  };

} // namespace vge
#endif
