// ############################################################################################
//  ImageBuilder1.cc
//  Implementation of CImageBuilder class (1st part)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "../3DEnvironment/FixedPointConsts.h"
#include "ImageBuilder.h"


namespace vge {

//
// Exported image helper functions
//
inline void _DitherColor_(uint8_t &c, uint8_t d)
{
  uint8_t in5 = (c >> 3);
  uint8_t in8 = in5 << 3;
  uint8_t del = (c - in8) << 5;

  if(del <= d) c = in5;
  else         c = (in5 >= 31) ? 31 : (in5 + 1);
}

void _OrderedDitherRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height)
{
  // Bayer matrix (for dithering RGB888 to RGB555)
  // NOTE: The matrix here was generated using the utility application,
  //       "util/minor/bym.cc" by using 255 as the scaling factor.
  const uint8_t msize       = 8;
  const uint8_t bayer[8][8] =
    { {   0, 194,  49, 243,  12, 206,  61, 255 },
      { 130,  65, 178, 113, 142,  77, 190, 125 },
      {  32, 227,  16, 210,  45, 239,  28, 223 },
      { 162,  97, 146,  81, 174, 109, 158,  93 },
      {   8, 202,  57, 251,   4, 198,  53, 247 },
      { 138,  73, 186, 121, 134,  69, 182, 117 },
      {  40, 235,  24, 219,  36, 231,  20, 215 },
      { 170, 105, 154,  89, 166, 101, 150,  85 } };

  for(uint32_t y = 0; y < height; y++) {
    for(uint32_t x = 0; x < width; x++) {
      uint8_t r = buffer->un8R;
      uint8_t g = buffer->un8G;
      uint8_t b = buffer->un8B;
      uint8_t d = bayer[y % msize][x % msize];
      _DitherColor_(r, d);
      _DitherColor_(g, d);
      _DitherColor_(b, d);
      buffer->un8R = r;
      buffer->un8G = g;
      buffer->un8B = b;
      buffer++;
    }
  }
}

void _ClipRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height)
{
  for(uint32_t i = 0; i < (width * height); i++) {
    buffer->un8R >>= 3;
    buffer->un8G >>= 3;
    buffer->un8B >>= 3;
    buffer++;
  }
}


//
// The image builder class
//
CImageBuilder::CImageBuilder()
: _ptBuffer(NULL), _un32Width(0), _un32Height(0)
{
  memset(&_tImgInfo, 0, sizeof(_tImgInfo));
}

CImageBuilder::~CImageBuilder() throw()
{
  if(_ptBuffer) free(_ptBuffer);
}

void CImageBuilder::FreeImage()
{
  if(_ptBuffer) free(_ptBuffer);
  _ptBuffer   = NULL;
  _un32Width  = 0;
  _un32Height = 0;
  memset(&_tImgInfo, 0, sizeof(_tImgInfo));
}


TUN8RGB *CImageBuilder::GetLineBuffer(uint32_t y)
{
  _if_unlikely(y >= _un32Height) VError(VERR_InvalidParameter, _NFL_, "The line index (y) must be smaller than the height of the image");
  return(&_ptBuffer[y * _un32Width]);
}

const TUN8RGB *CImageBuilder::GetLineBuffer(uint32_t y) const
{
  _if_unlikely(y >= _un32Height) VError(VERR_InvalidParameter, _NFL_, "The line index (y) must be smaller than the height of the image");
  return(&_ptBuffer[y * _un32Width]);
}


//
// Transfering functions
//
void CImageBuilder::CopyImage(CImageBuilder &csDstImage) const
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be copied (this CImageBuilder is not initialized)");

  // Resize the destination image if necessary
  if((_un32Width != csDstImage._un32Width) || (_un32Height != csDstImage._un32Height))
    csDstImage.CreateEmptyImage(_un32Width, _un32Height);

  // Copy image data
  memcpy(csDstImage._ptBuffer, _ptBuffer, _un32Width * _un32Height * 3);
  memcpy(&csDstImage._tImgInfo, &_tImgInfo, sizeof(_tImgInfo));
}

void CImageBuilder::TransferCoefToImage(CImage &csVImage) const
{
  // Temporary variables
  TUN8RGB  *ptTmp;
  uint8_t  *pun8Buff, *pun8Tmp;
  uint32_t  un32Cnt;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_UDCT_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_UDCT_RGB5_ image");

  // Destroy previous image, if any
  if(csVImage._pun8Buffer) abmfree(csVImage._pun8Buffer);
  csVImage._pun8Buffer = NULL;
  csVImage._un32Width  = 0;
  csVImage._un32Height = 0;
  csVImage._un32SizeSL = 0;

  // Allocate memory
  pun8Buff = reinterpret_cast<uint8_t*>(abmmalloc(_un32Width * _un32Height * 4));
  if(pun8Buff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");

  // Convert image
  ptTmp   = _ptBuffer;
  pun8Tmp = pun8Buff;
  for(un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    *pun8Tmp++ = (ptTmp->un8B & 0xF8);
    *pun8Tmp++ = (ptTmp->un8G & 0xF8);
    *pun8Tmp++ = (ptTmp->un8R & 0xF8);
    *pun8Tmp++ = 0;
    ptTmp++;
  }

  // Reinitialize the image
  csVImage._pun8Buffer = pun8Buff;
  csVImage._un32Width  = _un32Width;
  csVImage._un32Height = _un32Height;
  csVImage._un32SizeSL = _un32Width * 4;
}

void CImageBuilder::TransferToImage(CImage &csVImage) const
{
  // Temporary variables
  TUN8RGB  *ptTmp;
  uint8_t  *pun8Buff, *pun8Tmp;
  uint32_t  un32Cnt;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");

  // Destroy previous image, if any
  if(csVImage._pun8Buffer) abmfree(csVImage._pun8Buffer);
  csVImage._pun8Buffer = NULL;
  csVImage._un32Width  = 0;
  csVImage._un32Height = 0;
  csVImage._un32SizeSL = 0;

  // Allocate memory
  pun8Buff = reinterpret_cast<uint8_t*>(abmmalloc(_un32Width * _un32Height * 4));
  if(pun8Buff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");

  // Convert image
  ptTmp   = _ptBuffer;
  pun8Tmp = pun8Buff;
  for(un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    *pun8Tmp++ = (ptTmp->un8B << 3);
    *pun8Tmp++ = (ptTmp->un8G << 3);
    *pun8Tmp++ = (ptTmp->un8R << 3);
    *pun8Tmp++ = 0;
    ptTmp++;
  }

  // Reinitialize the image
  csVImage._pun8Buffer = pun8Buff;
  csVImage._un32Width  = _un32Width;
  csVImage._un32Height = _un32Height;
  csVImage._un32SizeSL = _un32Width * 4;
}

void CImageBuilder::TransferToAnimatedImage(CAnimatedImage &csVAnimatedImage,
                                            uint8_t un8SrcRowCount,
                                            uint8_t un8SrcColCount) const
{
  // Temporary variables
  TUN8RGB  *ptTmp;
  uint8_t  *pun8Buff, *pun8Tmp, un8R, un8C;
  uint32_t  un32Rs, un32Cs, un32Xs, un32Ys, un32X, un32Y;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if(!un8SrcRowCount || !un8SrcColCount)
    VError(VERR_InvalidParameter, _NFL_, "None of the 'un8SrcRowCount' nor the 'un8SrcColCount parameter' can be zero");
  if(_un32Height % un8SrcRowCount)
    VError(VERR_InvalidParameter, _NFL_, "Image height must be divisible by 'un8SrcRowCount'");
  if(_un32Width % un8SrcColCount)
    VError(VERR_InvalidParameter, _NFL_, "Image width must be divisible by 'un8SrcColCount'");

  // Destroy previous image, if any
  if(csVAnimatedImage._pun8BufferArray) abmfree(csVAnimatedImage._pun8BufferArray);
  csVAnimatedImage._pun8BufferArray = NULL;
  csVAnimatedImage._pun8BufferMax   = NULL;
  csVAnimatedImage._pun8Buffer      = NULL;
  csVAnimatedImage._un32Width       = 0;
  csVAnimatedImage._un32Height      = 0;
  csVAnimatedImage._un32TotFrame    = 0;
  csVAnimatedImage._un32CurFrame    = 0;
  csVAnimatedImage._un32FrameSize   = 0;
  csVAnimatedImage._un32StepDelay   = 0;
  csVAnimatedImage._n8Step          = 0;
  csVAnimatedImage._un32SizeSL      = 0;

  // Allocate memory
  pun8Buff = reinterpret_cast<uint8_t*>(abmmalloc(_un32Width * _un32Height * 4));
  if(pun8Buff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");

  // Convert image
  pun8Tmp = pun8Buff;
  un32Rs  = _un32Height / un8SrcRowCount;
  un32Cs  = _un32Width / un8SrcColCount;
  for(un8R = 0; un8R < un8SrcRowCount; un8R++) {
    un32Ys = un8R * un32Rs;
    for(un8C = 0; un8C < un8SrcColCount; un8C++) {
      un32Xs = un8C * un32Cs;
      for(un32Y = un32Ys; un32Y < (un32Ys + un32Rs); un32Y++) {
        ptTmp  = &_ptBuffer[un32Y * _un32Width + un32Xs];
        for(un32X = un32Xs; un32X < (un32Xs + un32Cs); un32X++) {
          *pun8Tmp++ = (ptTmp->un8B << 3);
          *pun8Tmp++ = (ptTmp->un8G << 3);
          *pun8Tmp++ = (ptTmp->un8R << 3);
          *pun8Tmp++ = 0;
          ptTmp++;
        }
      }
    }
  }

  // Reinitialize the image
  csVAnimatedImage._pun8BufferArray = pun8Buff;
  csVAnimatedImage._pun8Buffer      = csVAnimatedImage._pun8BufferArray;
  csVAnimatedImage._un32Width       = un32Cs;
  csVAnimatedImage._un32Height      = un32Rs;
  csVAnimatedImage._un32TotFrame    = un8SrcRowCount * un8SrcColCount;
  csVAnimatedImage._un32FrameSize   = un32Cs * un32Rs * 4;
  csVAnimatedImage._pun8BufferMax   = csVAnimatedImage._pun8BufferArray +
    (csVAnimatedImage._un32FrameSize *
    (csVAnimatedImage._un32TotFrame - 1));
  csVAnimatedImage._n8Step          = 1;
  csVAnimatedImage._un32SizeSL      = un32Cs * 4;
}

void CImageBuilder::TransferToTexture(CTexture &csVTexture) const
{
  // Temporary variables
  TUN8RGB  *ptTmp;
  uint8_t  *pun8Buff, *pun8Tmp;
  uint32_t  un32Cnt;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if((_un32Width !=  16) && (_un32Width !=  32) &&
     (_un32Width !=  64) && (_un32Width != 128) &&
     (_un32Width != 256) && (_un32Width != 512))
    VError(VERR_InvalidParameter, _NFL_, "Texture width must be one of 16/32/64/128/256/512 pixels");

  // Destroy previous image, if any
  if(csVTexture._pun8Buffer) abmfree(csVTexture._pun8Buffer);
  csVTexture._un32Width     = 0;
  csVTexture._un32Height    = 0;
  csVTexture._un32MaxX      = 0;
  csVTexture._un32MaxY      = 0;
  csVTexture._un32Mask      = 0;
  csVTexture._un32MultShift = 0;

  // Allocate memory
  pun8Buff = reinterpret_cast<uint8_t*>(abmmalloc(_un32Width * _un32Height * 4));
  if(pun8Buff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for texture buffer");

  // Convert image
  ptTmp   = _ptBuffer;
  pun8Tmp = pun8Buff;
  for(un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    *pun8Tmp++ = ptTmp->un8B;
    *pun8Tmp++ = ptTmp->un8G;
    *pun8Tmp++ = ptTmp->un8R;
    *pun8Tmp++ = 0;
    ptTmp++;
  }

  // Reinitialize the texture
  csVTexture._pun8Buffer    = pun8Buff;
  csVTexture._un32Width     = _un32Width;
  csVTexture._un32Height    = _un32Height;
  csVTexture._un32MaxX      = csVTexture._un32Width - 1;
  csVTexture._un32MaxY      = csVTexture._un32Height - 1;
  csVTexture._un32Mask      = _TFRAC_MASK_ * _un32Width;
  csVTexture._un32MultShift = 0;
  while(_un32Width > (1U << ++csVTexture._un32MultShift));
}

void CImageBuilder::TransferToAnimatedTexture(CAnimatedTexture &csVAnimatedTexture,
                                              uint8_t un8SrcRowCount,
                                              uint8_t un8SrcColCount) const
{
  // Temporary variables
  TUN8RGB  *ptTmp;
  uint8_t  *pun8Buff, *pun8Tmp, un8R, un8C;
  uint32_t  un32Rs, un32Cs, un32Xs, un32Ys, un32X, un32Y;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if((un8SrcRowCount * un8SrcColCount) <= 1)
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");
  if(_un32Height % un8SrcRowCount)
    VError(VERR_InvalidParameter, _NFL_, "Texture height must be divisible by un8SrcRowCount");
  if(_un32Width % un8SrcColCount)
    VError(VERR_InvalidParameter, _NFL_, "Texture width must be divisible by un8SrcColCount");

  // Calculate texture size
  un32Rs  = _un32Height / un8SrcRowCount;
  un32Cs  = _un32Width / un8SrcColCount;
  if((un32Cs !=  16) && (un32Cs !=  32) &&
     (un32Cs !=  64) && (un32Cs != 128) &&
     (un32Cs != 256) && (un32Cs != 512))
    VError(VERR_InvalidParameter, _NFL_, "Texture (width / un8SrcColCount) must be one of 16/32/64/128/256/512 pixels");

  // Destroy previous image, if any
  if(csVAnimatedTexture._pun8BufferArray) abmfree(csVAnimatedTexture._pun8BufferArray);
  csVAnimatedTexture._bAligned        = false;
  csVAnimatedTexture._pun8BufferArray = NULL;
  csVAnimatedTexture._pun8BufferMax   = NULL;
  csVAnimatedTexture._pun8Buffer      = NULL;
  csVAnimatedTexture._un32Width       = 0;
  csVAnimatedTexture._un32Height      = 0;
  csVAnimatedTexture._un32MaxX        = 0;
  csVAnimatedTexture._un32MaxY        = 0;
  csVAnimatedTexture._un32TotFrame    = 0;
  csVAnimatedTexture._un32CurFrame    = 0;
  csVAnimatedTexture._un32FrameSize   = 0;
  csVAnimatedTexture._un32StepDelay   = 0;
  csVAnimatedTexture._n8Step          = 0;
  csVAnimatedTexture._un32Mask        = 0;
  csVAnimatedTexture._un32MultShift   = 0;

  // Allocate memory
  pun8Buff = reinterpret_cast<uint8_t*>(abmmalloc(_un32Width * _un32Height * 4));
  if(pun8Buff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");

  // Convert image
  pun8Tmp = pun8Buff;
  for(un8R = 0; un8R < un8SrcRowCount; un8R++) {
    un32Ys = un8R * un32Rs;
    for(un8C = 0; un8C < un8SrcColCount; un8C++) {
      un32Xs = un8C * un32Cs;
      for(un32Y = un32Ys; un32Y < (un32Ys + un32Rs); un32Y++) {
        ptTmp  = &_ptBuffer[un32Y * _un32Width + un32Xs];
        for(un32X = un32Xs; un32X < (un32Xs + un32Cs); un32X++) {
          *pun8Tmp++ = ptTmp->un8B;
          *pun8Tmp++ = ptTmp->un8G;
          *pun8Tmp++ = ptTmp->un8R;
          *pun8Tmp++ = 0;
          ptTmp++;
        }
      }
    }
  }

  // Reinitialize the texture
  csVAnimatedTexture._pun8BufferArray = pun8Buff;
  csVAnimatedTexture._pun8Buffer      = csVAnimatedTexture._pun8BufferArray;
  csVAnimatedTexture._un32Width       = un32Cs;
  csVAnimatedTexture._un32Height      = un32Rs;
  csVAnimatedTexture._un32MaxX        = csVAnimatedTexture._un32Width - 1;
  csVAnimatedTexture._un32MaxY        = csVAnimatedTexture._un32Height - 1;
  csVAnimatedTexture._un32TotFrame    = un8SrcRowCount * un8SrcColCount;
  csVAnimatedTexture._un32FrameSize   = un32Cs * un32Rs * 4;
  csVAnimatedTexture._pun8BufferMax   = csVAnimatedTexture._pun8BufferArray +
    (csVAnimatedTexture._un32FrameSize *
    (csVAnimatedTexture._un32TotFrame - 1));
  csVAnimatedTexture._n8Step          = 1;
  csVAnimatedTexture._un32Mask        = _TFRAC_MASK_ * un32Cs;
  csVAnimatedTexture._un32MultShift   = 0;
  while(un32Cs > (1U << ++csVAnimatedTexture._un32MultShift));
}

} // namespace vge
