// ############################################################################################
//  ImageBuilder4.cc
//  Implementation of CImageBuilder class (4th part)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../VideoEnvironment/VideoEnvironment.h"
#include "ImageBuilder.h"
#include "MiscellaneousUtilities.h"


namespace vge {

//
// Spatial size for DCT
//
static const uint32_t un32Spatial = 8;


//
// DMT related functions
//
void CImageBuilder::_DMTGrabIMatrix_(uint32_t un32XStart, uint32_t un32YStart, float *pfRMatrix, float *pfGMatrix, float *pfBMatrix, uint8_t un8Dim)
{
  // Temporary variables
  uint32_t un32X, un32Y, un32XR, un32YR, un32SPos, un32DPos;

  // Grab pixels (in floats)
  un32XStart *= uint32_t(un8Dim);
  un32YStart *= uint32_t(un8Dim);
  for(un32Y = 0; un32Y < uint32_t(un8Dim); un32Y++) {
    un32YR = un32Y + un32YStart;
    for(un32X = 0; un32X < uint32_t(un8Dim); un32X++) {
      un32XR   = un32X + un32XStart;
      un32DPos = un32Y * uint32_t(un8Dim) + un32X;
      if((un32XR < _un32Width) && (un32YR < _un32Height)) {
        un32SPos = un32YR * _un32Width + un32XR;
        pfRMatrix[un32DPos] = float(_ptBuffer[un32SPos].un8R);
        pfGMatrix[un32DPos] = float(_ptBuffer[un32SPos].un8G);
        pfBMatrix[un32DPos] = float(_ptBuffer[un32SPos].un8B);
      }
      else {
        pfRMatrix[un32DPos] = 0.0f;
        pfGMatrix[un32DPos] = 0.0f;
        pfBMatrix[un32DPos] = 0.0f;
      }
    }
  }
}

void CImageBuilder::_DMTPutIMatrix_(uint32_t un32XStart, uint32_t un32YStart, const float *pfRMatrix, const float *pfGMatrix, const float *pfBMatrix, uint8_t un8Dim, bool bVRGBLimit)
{
  // Temporary variables
  uint32_t un32X, un32Y, un32XR, un32YR, un32SPos, un32DPos;

  // Put pixels (in VRGB)
  un32XStart *= uint32_t(un8Dim);
  un32YStart *= uint32_t(un8Dim);
  for(un32Y = 0; un32Y < uint32_t(un8Dim); un32Y++) {
    un32YR = un32Y + un32YStart;
    for(un32X = 0; un32X < uint32_t(un8Dim); un32X++) {
      un32XR   = un32X + un32XStart;
      un32DPos = un32Y * uint32_t(un8Dim) + un32X;
      if((un32XR < _un32Width) && (un32YR < _un32Height)) {
        un32SPos = un32YR * _un32Width + un32XR;
        if(bVRGBLimit) {
          _ptBuffer[un32SPos].un8R = (pfRMatrix[un32DPos] < 0) ? 0 : ((pfRMatrix[un32DPos] > 31) ? 31 : uint8_t(pfRMatrix[un32DPos]));
          _ptBuffer[un32SPos].un8G = (pfGMatrix[un32DPos] < 0) ? 0 : ((pfGMatrix[un32DPos] > 31) ? 31 : uint8_t(pfGMatrix[un32DPos]));
          _ptBuffer[un32SPos].un8B = (pfBMatrix[un32DPos] < 0) ? 0 : ((pfBMatrix[un32DPos] > 31) ? 31 : uint8_t(pfBMatrix[un32DPos]));
        }
        else {
          _ptBuffer[un32SPos].un8R = (pfRMatrix[un32DPos] < 0) ? 0 : ((pfRMatrix[un32DPos] > 255) ? 255 : uint8_t(pfRMatrix[un32DPos]));
          _ptBuffer[un32SPos].un8G = (pfGMatrix[un32DPos] < 0) ? 0 : ((pfGMatrix[un32DPos] > 255) ? 255 : uint8_t(pfGMatrix[un32DPos]));
          _ptBuffer[un32SPos].un8B = (pfBMatrix[un32DPos] < 0) ? 0 : ((pfBMatrix[un32DPos] > 255) ? 255 : uint8_t(pfBMatrix[un32DPos]));
        }
      }
    }
  }
}

void CImageBuilder::_PointSmooth_(uint32_t un32X, uint32_t un32Y, uint32_t un32Radius)
{
  // Temporary variables
  int32_t  n32X, n32Y;
  uint32_t un32R, un32G, un32B, un32C, un32Pos;

  // Grab pixels
  un32R = un32G = un32B = un32C = 0;
  for(n32X = (int32_t(un32X) - int32_t(un32Radius)); n32X < (int32_t(un32X) + int32_t(un32Radius)); n32X++) {
    for(n32Y = (int32_t(un32Y) - int32_t(un32Radius)); n32Y < (int32_t(un32Y) + int32_t(un32Radius)); n32Y++) {
      if((n32X > 0) && (n32Y > 0) &&
      (n32X < int32_t(_un32Width)) && (n32Y < int32_t(_un32Height))) {
        un32Pos = uint32_t(n32Y) * _un32Width + uint32_t(n32X);
        un32R += _ptBuffer[un32Pos].un8R;
        un32G += _ptBuffer[un32Pos].un8G;
        un32B += _ptBuffer[un32Pos].un8B;
        un32C++;
      }
    }
  }

  // Store the averaged result
  if(un32C) {
    un32Pos = un32Y * _un32Width + un32X;
    _ptBuffer[un32Pos].un8R = uint8_t(un32R / un32C);
    _ptBuffer[un32Pos].un8G = uint8_t(un32G / un32C);
    _ptBuffer[un32Pos].un8B = uint8_t(un32B / un32C);
  }
}

void CImageBuilder::_PointSmooth2_(uint32_t un32X, uint32_t un32Y)
{
  // Temporary variables
  int32_t  n32X, n32Y;
  uint32_t un32R, un32G, un32B, un32C, un32Pos;

  // Grab pixels
  un32R = un32G = un32B = un32C = 0;
  for(n32X = int32_t(un32X) - 2; n32X < (int32_t(un32X) + 2); n32X++) {
    for(n32Y = int32_t(un32Y) - 2; n32Y < (int32_t(un32Y) + 2); n32Y++) {
      if((n32X > 0) && (n32Y > 0) &&
      (n32X < int32_t(_un32Width)) && (n32Y < int32_t(_un32Height))) {
        un32Pos = uint32_t(n32Y) * _un32Width + uint32_t(n32X);
        un32R += _ptBuffer[un32Pos].un8R;
        un32G += _ptBuffer[un32Pos].un8G;
        un32B += _ptBuffer[un32Pos].un8B;
        un32C++;
      }
    }
  }

  // Store the averaged result
  if(un32C) {
    un32Pos = un32Y * _un32Width + un32X;
    _ptBuffer[un32Pos].un8R = uint8_t(un32R / un32C);
    _ptBuffer[un32Pos].un8G = uint8_t(un32G / un32C);
    _ptBuffer[un32Pos].un8B = uint8_t(un32B / un32C);
  }
}


//
// DCT related functions
//
void CImageBuilder::PerformDCT(CImageBuilder &csDstImage, uint8_t un8ReduceCoefficientStrength)
{
  // Temporary variables
  uint32_t un32XC, un32YC, un32XM, un32YM, un32CRLimit, un32CRRetained;
  float    aafK[un32Spatial][un32Spatial];
  float    aafR[un32Spatial][un32Spatial], aafG[un32Spatial][un32Spatial], aafB[un32Spatial][un32Spatial];

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_RGB5_ image");
  if((uint32_t(_un32Width / un32Spatial) * un32Spatial) != _un32Width)
    VError(VERR_InvalidParameter, _NFL_, "Image width must be divisible by %d for DCT transformation", un32Spatial);
  if((uint32_t(_un32Height / un32Spatial) * un32Spatial) != _un32Height)
    VError(VERR_InvalidParameter, _NFL_, "Image height must be divisible by %d for DCT transformation", un32Spatial);

  // Resize the destination image if necessary
  if((_un32Width != csDstImage._un32Width) || (_un32Height != csDstImage._un32Height))
    csDstImage.CreateEmptyImage(_un32Width, _un32Height);

  // Create the DCT kernel
  VDMTCreateDCTKernelMatrix(&aafK[0][0], un32Spatial);
  VDMTDumpMatrix("Dumped DCT kernel matrix :", &aafK[0][0], un32Spatial);

  // Do the transformation
  un32XM = _un32Width / un32Spatial;  un32XM = ((un32XM * un32Spatial) < _un32Width)  ? (un32XM + 1) : un32XM;
  un32YM = _un32Height / un32Spatial; un32YM = ((un32YM * un32Spatial) < _un32Height) ? (un32YM + 1) : un32YM;
  for(un32YC = 0; un32YC < un32YM; un32YC++) {
    for(un32XC = 0; un32XC < un32XM; un32XC++) {
      _DMTGrabIMatrix_(un32XC, un32YC, &aafR[0][0], &aafG[0][0], &aafB[0][0], un32Spatial);
      VDMTMultMatricesKIK(&aafR[0][0], &aafK[0][0], un32Spatial);
      VDMTMultMatricesKIK(&aafG[0][0], &aafK[0][0], un32Spatial);
      VDMTMultMatricesKIK(&aafB[0][0], &aafK[0][0], un32Spatial);
      VDMTQuantitizeIMatrix(&aafR[0][0], un32Spatial, un8ReduceCoefficientStrength, &un32CRLimit, &un32CRRetained);
      VDMTQuantitizeIMatrix(&aafG[0][0], un32Spatial, un8ReduceCoefficientStrength);
      VDMTQuantitizeIMatrix(&aafB[0][0], un32Spatial, un8ReduceCoefficientStrength);
      csDstImage._DMTPutIMatrix_(un32XC, un32YC, &aafR[0][0], &aafG[0][0], &aafB[0][0], un32Spatial, false);
    }
  }

  // Display and set some information
  csDstImage._tImgInfo.un8Flag        = _IMG_TYPE_UDCT_RGB5_;
  csDstImage._tImgInfo.un32OW         = _un32Width;
  csDstImage._tImgInfo.un32OH         = _un32Height;
  csDstImage._tImgInfo.un32Spatial    = un32Spatial;
  csDstImage._tImgInfo.un32CRLimit    = un32CRLimit;
  csDstImage._tImgInfo.un32CRRetained = un32CRRetained;
  VInfo(_NFL_, "DCT coefficients retained : %02d/%02d (%05.2f%%)",
        un32CRRetained, un32Spatial * un32Spatial,
        float(un32CRRetained) / float(un32Spatial * un32Spatial) * 100.0f);
}

void CImageBuilder::PerformInvDCT(CImageBuilder &csDstImage, bool bSmoothDCPoints)
{
  // Temporary variables
  uint32_t un32XC, un32YC, un32XM, un32YM;
  float    aafK[un32Spatial][un32Spatial];
  float    aafR[un32Spatial][un32Spatial], aafG[un32Spatial][un32Spatial], aafB[un32Spatial][un32Spatial];

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_UDCT_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_UDCT_RGB5_ image");

  // Resize the destination image if necessary
  if((_un32Width != csDstImage._un32Width) || (_un32Height != csDstImage._un32Height))
    csDstImage.CreateEmptyImage(_un32Width, _un32Height);

  // Create the IDCT kernel
  VDMTCreateInvDCTKernelMatrix(&aafK[0][0], un32Spatial);
  VDMTDumpMatrix("Dumped IDCT kernel matrix :", &aafK[0][0], un32Spatial);

  // Do the transformation
  un32XM = _un32Width / un32Spatial;  un32XM = ((un32XM * un32Spatial) < _un32Width)  ? (un32XM + 1) : un32XM;
  un32YM = _un32Height / un32Spatial; un32YM = ((un32YM * un32Spatial) < _un32Height) ? (un32YM + 1) : un32YM;
  for(un32YC = 0; un32YC < un32YM; un32YC++) {
    for(un32XC = 0; un32XC < un32XM; un32XC++) {
      _DMTGrabIMatrix_(un32XC, un32YC, &aafR[0][0], &aafG[0][0], &aafB[0][0], un32Spatial);
      VDMTDequantitizeIMatrix(&aafR[0][0], un32Spatial);
      VDMTDequantitizeIMatrix(&aafG[0][0], un32Spatial);
      VDMTDequantitizeIMatrix(&aafB[0][0], un32Spatial);
      VDMTMultMatricesKIK(&aafR[0][0], &aafK[0][0], un32Spatial);
      VDMTMultMatricesKIK(&aafG[0][0], &aafK[0][0], un32Spatial);
      VDMTMultMatricesKIK(&aafB[0][0], &aafK[0][0], un32Spatial);
      csDstImage._DMTPutIMatrix_(un32XC, un32YC, &aafR[0][0], &aafG[0][0], &aafB[0][0], un32Spatial, true);
    }
  }
  memset(&csDstImage._tImgInfo, 0, sizeof(csDstImage._tImgInfo));

  // Smooth the DC points
  if(bSmoothDCPoints) {
    for(un32YC = 0; un32YC < un32YM; un32YC++)
      for(un32XC = 0; un32XC < un32XM; un32XC++)
        csDstImage._PointSmooth2_(un32XC * un32Spatial, un32YC * un32Spatial);
  }
}

void CImageBuilder::PerformDCTImageCompaction()
{
  // Temporary variables
  TUN8RGB *ptBuf;
  uint32_t      un32BSize, un32Ys, un32Xs, un32R, un32C;
  uint32_t      un32Y1, un32X1, un32Y2, un32X2, un32RPos, un32WPos;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_UDCT_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_UDCT_RGB5_ image");

  // Allocate temporary memory
  un32BSize = (_un32Width / _tImgInfo.un32Spatial) * (_un32Height / _tImgInfo.un32Spatial) *
              (_tImgInfo.un32CRRetained + 2);
  ptBuf = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * un32BSize));
  memset(ptBuf, 0, sizeof(TUN8RGB) * un32BSize);
  if(ptBuf == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for temporary compaction buffer");

  // Do compaction
  un32WPos = 0;
  for(un32Ys = 0; un32Ys < (_un32Height / _tImgInfo.un32Spatial); un32Ys++) {
    // Calculate the starting Y coordinate
    un32Y1 = un32Ys * _tImgInfo.un32Spatial;
    // Loop for the horizontal spatials
    for(un32Xs = 0; un32Xs < (_un32Width / _tImgInfo.un32Spatial); un32Xs++) {
      // Calculate the starting X coordinate
      un32X1 = un32Xs * _tImgInfo.un32Spatial;
      // Process a spatial
      for(un32R = 0; un32R < _tImgInfo.un32Spatial; un32R++) {
        // Calculate the real Y coordinate
        un32Y2 = un32Y1 + un32R;
        for(un32C = 0; un32C < _tImgInfo.un32Spatial; un32C++) {
          if((un32R * un32C) < _tImgInfo.un32CRLimit) {
            // Calculate the real X coordinate
            un32X2 = un32X1 + un32C;
            // Calculate read address
            un32RPos = un32Y2 * _un32Width + un32X2;
            // Copy data to compaction buffer
            ptBuf[un32WPos].un8R = _ptBuffer[un32RPos].un8R;
            ptBuf[un32WPos].un8G = _ptBuffer[un32RPos].un8G;
            ptBuf[un32WPos].un8B = _ptBuffer[un32RPos].un8B;
            un32WPos++;
          }
        }
      }
      // Save the maxima
      un32RPos = (un32Y1 + 7) * _un32Width + (un32X1 + 6);
      ptBuf[un32WPos].un8R = _ptBuffer[un32RPos].un8R;
      ptBuf[un32WPos].un8G = _ptBuffer[un32RPos].un8G;
      ptBuf[un32WPos].un8B = _ptBuffer[un32RPos].un8B;
      un32WPos++; un32RPos++;
      ptBuf[un32WPos].un8R = _ptBuffer[un32RPos].un8R;
      ptBuf[un32WPos].un8G = _ptBuffer[un32RPos].un8G;
      ptBuf[un32WPos].un8B = _ptBuffer[un32RPos].un8B;
      un32WPos++;
    }
  }

  // Update variables
  free(_ptBuffer);
  _ptBuffer         = ptBuf;
  _un32Width        = un32WPos;
  _un32Height       = 1;
  _tImgInfo.un8Flag = _IMG_TYPE_CDCT_RGB5_;
}

void CImageBuilder::PerformDCTImageExpansion()
{
  // Temporary variables
  TUN8RGB *ptBuf;
  uint32_t      un32BSize, un32Ys, un32Xs, un32R, un32C;
  uint32_t      un32Y1, un32X1, un32Y2, un32X2, un32RPos, un32WPos;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_CDCT_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_CDCT_RGB5_ image");

  // Allocate temporary memory
  un32BSize = _tImgInfo.un32OW * _tImgInfo.un32OH;
  ptBuf = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * un32BSize));
  memset(ptBuf, 127, sizeof(TUN8RGB) * un32BSize);
  if(ptBuf == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for temporaty expansion buffer");

  // Do expansion
  un32RPos = 0;
  for(un32Ys = 0; un32Ys < (_tImgInfo.un32OH / _tImgInfo.un32Spatial); un32Ys++) {
    // Calculate the starting Y coordinate
    un32Y1 = un32Ys * _tImgInfo.un32Spatial;
    // Loop for the horizontal spatials
    for(un32Xs = 0; un32Xs < (_tImgInfo.un32OW / _tImgInfo.un32Spatial); un32Xs++) {
      // Calculate the starting X coordinate
      un32X1 = un32Xs * _tImgInfo.un32Spatial;
      // Process a spatial
      for(un32R = 0; un32R < _tImgInfo.un32Spatial; un32R++) {
        // Calculate the real Y coordinate
        un32Y2 = un32Y1 + un32R;
        for(un32C = 0; un32C < _tImgInfo.un32Spatial; un32C++) {
          if((un32R * un32C) < _tImgInfo.un32CRLimit) {
            // Calculate the real X coordinate
            un32X2 = un32X1 + un32C;
            // Calculate write address
            un32WPos = un32Y2 * _tImgInfo.un32OW + un32X2;
            // Copy data to compaction buffer
            ptBuf[un32WPos].un8R = _ptBuffer[un32RPos].un8R;
            ptBuf[un32WPos].un8G = _ptBuffer[un32RPos].un8G;
            ptBuf[un32WPos].un8B = _ptBuffer[un32RPos].un8B;
            un32RPos++;
          }
        }
      }
      // Get the maxima
      un32WPos = (un32Y1 + 7) * _tImgInfo.un32OW + (un32X1 + 6);
      ptBuf[un32WPos].un8R = _ptBuffer[un32RPos].un8R;
      ptBuf[un32WPos].un8G = _ptBuffer[un32RPos].un8G;
      ptBuf[un32WPos].un8B = _ptBuffer[un32RPos].un8B;
      un32RPos++; un32WPos++;
      ptBuf[un32WPos].un8R = _ptBuffer[un32RPos].un8R;
      ptBuf[un32WPos].un8G = _ptBuffer[un32RPos].un8G;
      ptBuf[un32WPos].un8B = _ptBuffer[un32RPos].un8B;
      un32RPos++;
    }
  }

  // Update variables
  free(_ptBuffer);
  _ptBuffer         = ptBuf;
  _un32Width        = _tImgInfo.un32OW;
  _un32Height       = _tImgInfo.un32OH;
  _tImgInfo.un8Flag = _IMG_TYPE_UDCT_RGB5_;
}


//
// YUV related functions
//
void CImageBuilder::ConvertRGB2YUV()
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  uint8_t  un8R, un8G, un8B, un8Y, un8U, un8V;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_RGB5_ image");

  // Do conversion
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get the pixel
      un8R = _ptBuffer[un32Pos2].un8R;
      un8G = _ptBuffer[un32Pos2].un8G;
      un8B = _ptBuffer[un32Pos2].un8B;
      // Convert to YUV
      V_RGB2YUV(un8R, un8G, un8B, un8Y, un8U, un8V);
      // Put the pixel back
      _ptBuffer[un32Pos2].un8R = un8Y;
      _ptBuffer[un32Pos2].un8G = un8U;
      _ptBuffer[un32Pos2].un8B = un8V;
    }
  }

  // Update variables
  _tImgInfo.un8Flag = _IMG_TYPE_FYUV_;
}

void CImageBuilder::ConvertYUV2RGB()
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  uint8_t  un8R, un8G, un8B, un8Y, un8U, un8V;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_FYUV_ image");

  // Do conversion
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get the pixel
      un8Y = _ptBuffer[un32Pos2].un8R;
      un8U = _ptBuffer[un32Pos2].un8G;
      un8V = _ptBuffer[un32Pos2].un8B;
      // Convert to YUV
      V_YUV2RGB(un8Y, un8U, un8V, un8R, un8G, un8B);
      // Put the pixel back
      _ptBuffer[un32Pos2].un8R = un8R;
      _ptBuffer[un32Pos2].un8G = un8G;
      _ptBuffer[un32Pos2].un8B = un8B;
    }
  }

  // Update variables
  _tImgInfo.un8Flag = _IMG_TYPE_RGB5_;
}

void CImageBuilder::ConvertYUV2YV12()
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_FYUV_ image");

  // Calculate the size of the U & V buffer
  uint32_t un32W = (_un32Width + 1) / 2;
  uint32_t un32H = (_un32Height + 1) / 2;

  // Clone and resize the image
  CImageBuilder *pcsImg;
  CloneImage(&pcsImg);
  pcsImg->Resize(un32W, un32H);

  // Calculate pointers
  u_char *pucY = reinterpret_cast<u_char*>(_ptBuffer);
  u_char *pucU = pucY + (_un32Width * _un32Height);
  u_char *pucV = pucU + (un32W * un32H);
  u_char *pucE = pucV + (un32W * un32H);

  // Grab the Y data (unscaled from the original image)
  for(uint32_t un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    *pucY++ = _ptBuffer[un32Cnt].un8R;
  }

  // Grab the U & Y data (scaled from the clone image)
  for(uint32_t un32Cnt = 0; un32Cnt < (un32W * un32H); un32Cnt++) {
    *pucU++ = pcsImg->_ptBuffer[un32Cnt].un8G;
    *pucV++ = pcsImg->_ptBuffer[un32Cnt].un8B;
  }

  // Clear the unused part of the buffer
  memset(pucE, 0, (_un32Width * _un32Height * 2) - (un32W * un32H * 2));

  // Adjust the image flags
  _tImgInfo.un8Flag = _IMG_TYPE_YV12_;
  _tImgInfo.un32OW  = (_un32Width * _un32Height) + (un32W * un32H * 2);

  // Delete the clone image
  delete pcsImg;
}

void CImageBuilder::ConvertYV122YUV()
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_YV12_)
    VError(VERR_InvalidCall, _NFL_, "Source image must be an _IMG_TYPE_YV12_ image");

  // Calculate the size of the U & V buffer
  uint32_t un32W = (_un32Width + 1) / 2;
  uint32_t un32H = (_un32Height + 1) / 2;

  // More check
  vassert(_tImgInfo.un32OW == (_un32Width*_un32Height + un32W*un32H*2),
          "This _IMG_TYPE_YV12_ image is corrupted");

  // Create an empty image for resizing
  CImageBuilder csImg;
  csImg.CreateEmptyImage(un32W, un32H);

  // Calculate pointers
  u_char *pucY = reinterpret_cast<u_char*>(_ptBuffer);
  u_char *pucU = pucY + (_un32Width * _un32Height);
  u_char *pucV = pucU + (un32W * un32H);

  // Grab the U & Y data and scale back to the original size
  for(uint32_t un32Cnt = 0; un32Cnt < (un32W * un32H); un32Cnt++) {
    csImg._ptBuffer[un32Cnt].un8G = *pucU++;
    csImg._ptBuffer[un32Cnt].un8B = *pucV++;
  }
  csImg.Resize(_un32Width, _un32Height);

  // Clone the scaled image for transferring
  CImageBuilder *pcsDst;
  csImg.CloneImage(&pcsDst);

  // Grab the Y data
  for(uint32_t un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    pcsDst->_ptBuffer[un32Cnt].un8R = *pucY++;
  }

  // Adjust the image data
  memcpy(_ptBuffer, pcsDst->_ptBuffer, _un32Width * _un32Height * 3);
  _tImgInfo.un8Flag = _IMG_TYPE_FYUV_;
  _tImgInfo.un32OW  = 0;

  // Delete the clone image
  delete pcsDst;
}

void CImageBuilder::YUVShiftY(int8_t n8ShiftFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8Y;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8Y = int8_t(_ptBuffer[un32Pos2].un8R) + n8ShiftFactor;
      n8Y = (n8Y >= 0) ? ((n8Y <= MAX_COLOR_VALUE) ? n8Y : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8R = uint8_t(n8Y);
    }
  }
}

void CImageBuilder::YUVShiftU(int8_t n8ShiftFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8U;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8U = int8_t(_ptBuffer[un32Pos2].un8G) + n8ShiftFactor;
      n8U = (n8U >= 0) ? ((n8U <= MAX_COLOR_VALUE) ? n8U : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8G = uint8_t(n8U);
    }
  }
}

void CImageBuilder::YUVShiftV(int8_t n8ShiftFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8V;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8V = int8_t(_ptBuffer[un32Pos2].un8B) + n8ShiftFactor;
      n8V = (n8V >= 0) ? ((n8V <= MAX_COLOR_VALUE) ? n8V : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8B = uint8_t(n8V);
    }
  }
}

void CImageBuilder::YUVScaleY(float fScaleFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8Y;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8Y = int8_t(_ptBuffer[un32Pos2].un8R * fScaleFactor);
      n8Y = (n8Y >= 0) ? ((n8Y <= MAX_COLOR_VALUE) ? n8Y : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8R = uint8_t(n8Y);
    }
  }
}

void CImageBuilder::YUVScaleU(float fScaleFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8U;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8U = int8_t((int8_t(_ptBuffer[un32Pos2].un8G) - 16) * fScaleFactor) + 16;
      n8U = (n8U >= 0) ? ((n8U <= MAX_COLOR_VALUE) ? n8U : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8G = uint8_t(n8U);
    }
  }
}

void CImageBuilder::YUVScaleV(float fScaleFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8V;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8V = int8_t((int8_t(_ptBuffer[un32Pos2].un8B) - 16) * fScaleFactor) + 16;
      n8V = (n8V >= 0) ? ((n8V <= MAX_COLOR_VALUE) ? n8V : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8B = uint8_t(n8V);
    }
  }
}

void CImageBuilder::YUVScaleUV(float fUScaleFactor, float fVScaleFactor)
{
  // Temporary variables
  uint32_t un32CntX, un32CntY, un32Pos, un32Pos2;
  int8_t   n8U, n8V;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_FYUV_)
    VError(VERR_InvalidCall, _NFL_, "Can only process an _IMG_TYPE_FYUV_ image");

  // Do filter
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32Pos = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      un32Pos2 = un32Pos + un32CntX;
      // Get and process the pixel
      n8U = int8_t((int8_t(_ptBuffer[un32Pos2].un8G) - 16) * fUScaleFactor) + 16;
      n8U = (n8U >= 0) ? ((n8U <= MAX_COLOR_VALUE) ? n8U : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8G = uint8_t(n8U);
      n8V = int8_t((int8_t(_ptBuffer[un32Pos2].un8B) - 16) * fVScaleFactor) + 16;
      n8V = (n8V >= 0) ? ((n8V <= MAX_COLOR_VALUE) ? n8V : MAX_COLOR_VALUE) : 0;
      _ptBuffer[un32Pos2].un8B = uint8_t(n8V);
    }
  }
}

} // namespace vge
