// ############################################################################################
//  MiscellaneousUtilities.cc
//  Implementation of various mathematic utilities related to image processing
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "MiscellaneousUtilities.h"


namespace vge {

//
// General functions
//
void VDMTDumpMatrix(const char *pcInfoText, const float *pfMatrix, uint8_t un8Dim)
{
  // Check parameters
  if((pfMatrix == NULL) || (un8Dim < 2) || (un8Dim > 64))
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");

  // Dump matrix
  VInfo(_NFL_, pcInfoText); VPrint("  ");
  for(size_t i = 0; i < un8Dim; ++i) {
    for(size_t j = 0; j < un8Dim; ++j) {
      VPrint("%+04.2f ", pfMatrix[i * un8Dim + j]);
    }
    VPrint("\n  ");
  }
}

void VDMTTransposeMatrix(float *pfMatrix, uint8_t un8Dim)
{
  // Check parameters
  if((pfMatrix == NULL) || (un8Dim < 2) || (un8Dim > 64))
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");

  // Allocate temporary storage
  float *pfTM = _nothrow_new float[size_t(un8Dim * un8Dim)];
  if(pfTM == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for allocating temporary matrix storage");

  // Transpose matrix
  for(size_t r = 0; r < un8Dim; ++r)
    for(size_t c = 0; c < un8Dim; ++c)
      pfTM[c * un8Dim + r] = pfMatrix[r * un8Dim + c];

  // Copy result
  memcpy(pfMatrix, pfTM, sizeof(float) * size_t(un8Dim * un8Dim));
  delete pfTM;

}

void VDMTMultMatricesKIK(float *pfIMatrix, const float *pfKMatrix, uint8_t un8Dim)
{
  // Check parameters
  if((pfIMatrix == NULL) || (pfKMatrix == NULL) || (un8Dim < 2) || (un8Dim > 64))
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");

  // Allocate temporary storage
  float *pfTM = _nothrow_new float[uint32_t(un8Dim * un8Dim)];
  if(pfTM == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for allocating temporary matrix storage");

  // Multiply matrices (1st step)
  for(size_t r = 0; r < un8Dim; ++r) {
    for(size_t c = 0; c < un8Dim; ++c) {
      float fSum = 0.0f;
      for(size_t i = 0; i < un8Dim; ++i) {
        fSum += (pfKMatrix[r * un8Dim + i] * pfIMatrix[i * un8Dim + c]);
      }
      pfTM[r * un8Dim + c] = fSum;
    }
  }

  // Copy the intermediate result
  memcpy(pfIMatrix, pfTM, sizeof(float) * size_t(un8Dim * un8Dim));

  // Multiply matrices (2nd step)
  for(size_t r = 0; r < un8Dim; ++r) {
    for(size_t c = 0; c < un8Dim; ++c) {
      float fSum = 0.0f;
      for(size_t i = 0; i < un8Dim; ++i) {
        fSum += (pfIMatrix[r * un8Dim + i] * pfKMatrix[i * un8Dim + c]);
      }
      pfTM[r * un8Dim + c] = fSum;
    }
  }

  // Copy the final result
  memcpy(pfIMatrix, pfTM, sizeof(float) * size_t(un8Dim * un8Dim));

  // Free memory
  delete pfTM;
}


//
// Quantitization functions
//
void VDMTQuantitizeIMatrix(float *pfIMatrix, uint8_t un8Dim, uint8_t un8CRStrength,
                           uint32_t *pun32CRLimit, uint32_t *pun32CRRetained)
{
  // Check parameters
  if((pfIMatrix == NULL) || (un8Dim < 2) || (un8Dim > 64))
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");

  // Find the maximum value
  float fMax = 0.0f;
  for(size_t r = 0; r < un8Dim; ++r) {
    for(size_t c = 0; c < un8Dim; ++c) {
      float fCur = fabs(pfIMatrix[r * un8Dim + c]);
      if(fCur > fMax) fMax = fCur;
    }
  }
  if(fMax == 0.0f) fMax = 1.0f;

  // Discard some coefficients
  size_t crLimit;
  size_t crRetained;
  if(un8CRStrength > 0) {
    crLimit    = size_t(un8Dim * un8Dim) >> un8CRStrength;
    crRetained = 0;
    for(size_t r = 0; r < un8Dim; ++r) {
      for(size_t c = 0; c < un8Dim; ++c) {
        if(r*c >= crLimit)
          pfIMatrix[r * un8Dim + c] = 0.0f;
        else
          ++crRetained;
      }
    }
  }
  else {
    crLimit    = size_t(un8Dim * un8Dim);
    crRetained = size_t(un8Dim * un8Dim - 2);
  }
  if(pun32CRLimit)    *pun32CRLimit    = crLimit;
  if(pun32CRRetained) *pun32CRRetained = crRetained;

  // Quantitize the values
  for(size_t r = 0; r < un8Dim; ++r) {
    for(size_t c = 0; c < un8Dim; ++c) {
      float fCur = pfIMatrix[r * un8Dim + c] / fMax;
      pfIMatrix[r * un8Dim + c] = (fCur + 1.0f) / 2.0f * 255.0f;
    }
  }
  uint16_t un16Max = uint16_t(fMax);
  uint8_t *pun8Max = reinterpret_cast<uint8_t*>(&un16Max);
  pfIMatrix[uint32_t(un8Dim * un8Dim) - 2] = float(*(pun8Max + 0));
  pfIMatrix[uint32_t(un8Dim * un8Dim) - 1] = float(*(pun8Max + 1));
}

void VDMTDequantitizeIMatrix(float *pfIMatrix, uint8_t un8Dim)
{
  // Check parameters
  if((pfIMatrix == NULL) || (un8Dim < 2) || (un8Dim > 64))
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");

  // Get the maximum value
  uint16_t un16Max;
  uint8_t *pun8Max = reinterpret_cast<uint8_t*>(&un16Max);
  *(pun8Max + 0) = uint8_t(pfIMatrix[uint32_t(un8Dim * un8Dim) - 2]);
  *(pun8Max + 1) = uint8_t(pfIMatrix[uint32_t(un8Dim * un8Dim) - 1]);

  // Dequantitize values
  for(size_t r = 0; r < un8Dim; ++r) {
    for(size_t c = 0; c < un8Dim; ++c) {
      float fCur = pfIMatrix[r * un8Dim + c];
      pfIMatrix[r * un8Dim + c] = (fCur * 2.0f / 255.0f - 1.0f) * float(un16Max);
    }
  }
  pfIMatrix[size_t(un8Dim * un8Dim) - 2] = 0.0f;
  pfIMatrix[size_t(un8Dim * un8Dim) - 1] = 0.0f;
}


//
// DCT related functions
//
void VDMTCreateDCTKernelMatrix(float *pfKMatrix, uint8_t un8Dim)
{
  // Check parameters
  if((pfKMatrix == NULL) || (un8Dim < 2) || (un8Dim > 64))
    VError(VERR_InvalidParameter, _NFL_, "Invalid parameter(s)");

  // Fill matrix
  float f1 = float(sqrt(float(1.0) / float(un8Dim)));
  float f2 = float(sqrt(float(2.0) / float(un8Dim)));
  float fD = 2.0f * float(un8Dim);
  for(size_t k = 0; k < un8Dim; ++k) {
    for(size_t j = 0; j < un8Dim; ++j) {
      if(!k)
        pfKMatrix[k * un8Dim + j] = f1;
      else
        pfKMatrix[k * un8Dim + j] = f2 * cos(float(3.141592654) * (2.0f * float(j) + 1) * float(k) / fD);
    }
  }
}

void VDMTCreateInvDCTKernelMatrix(float *pfKMatrix, uint8_t un8Dim)
{
  VDMTCreateDCTKernelMatrix(pfKMatrix, un8Dim);
  VDMTTransposeMatrix(pfKMatrix, un8Dim);
}

} // namespace vge
