// ############################################################################################
//  CameraControl.cc
//  Implementation of CWalkingCameraControl class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/LookUpTables.h"
#include "3DEnvironment.h"


namespace vge {

//
// The control class for 1st person view camera
//
CWalkingCameraControl::CWalkingCameraControl(PFNVortexWCCCallback pfnWCCCalback, void *pvUserData)
: _pfnCallback(pfnWCCCalback), _pvUserCBData(pvUserData), _pcsSkybox(NULL)
{
  ResetAll();
}

CWalkingCameraControl::~CWalkingCameraControl() throw()
{
  delete _pcsSkybox;
}

void CWalkingCameraControl::SetCallback(PFNVortexWCCCallback pfnWCCCalback, void *pvUserData)
{
  _pfnCallback  = pfnWCCCalback;
  _pvUserCBData = pvUserData;
}

void CWalkingCameraControl::StepForward(float fAmount)
{
  // Calculate the new position
  float fTrans[3];
  fTrans[0] = fAmount * (_SIN_(_an32RotXY[1]));
  fTrans[1] = 0;
  fTrans[2] = fAmount * (_COS_(_an32RotXY[1]));

  // Calculate collision
  bool bOK;
  if(_pfnCallback) 
    bOK = _pfnCallback(EWCC_Step, this, _afPosXYZ, fTrans, _pvUserCBData);
  else
    bOK = CameraCallback(EWCC_Step, _afPosXYZ, fTrans, _pvUserCBData);

  // Update only if allowed by the user callback function
  if(bOK) {
    _afPosXYZ[0] += fTrans[0];
    _afPosXYZ[2] += fTrans[2];
  }
}

void CWalkingCameraControl::StepBackward(float fAmount)
{
  // Calculate the new position
  float fTrans[3];
  fTrans[0] = -fAmount * (_SIN_(_an32RotXY[1]));
  fTrans[1] = 0;
  fTrans[2] = -fAmount * (_COS_(_an32RotXY[1]));

  // Calculate collision
  bool bOK;
  if(_pfnCallback) 
    bOK = _pfnCallback(EWCC_Step, this, _afPosXYZ, fTrans, _pvUserCBData);
  else
    bOK = CameraCallback(EWCC_Step, _afPosXYZ, fTrans, _pvUserCBData);

  // Update only if allowed by the user callback function
  if(bOK) {
    _afPosXYZ[0] += fTrans[0];
    _afPosXYZ[2] += fTrans[2];
  }
}

void CWalkingCameraControl::StrafeLeft(float fAmount)
{
  // Calculate the new position
  float fTrans[3];
  fTrans[0] = -fAmount * (COS(_an32RotXY[1]));
  fTrans[1] = 0;
  fTrans[2] = fAmount * (SIN(_an32RotXY[1]));

  // Calculate collision
  bool bOK;
  if(_pfnCallback) 
    bOK = _pfnCallback(EWCC_Step, this, _afPosXYZ, fTrans, _pvUserCBData);
  else
    bOK = CameraCallback(EWCC_Step, _afPosXYZ, fTrans, _pvUserCBData);

  // Update only if allowed by the user callback function
  if(bOK) {
    _afPosXYZ[0] += fTrans[0];
    _afPosXYZ[2] += fTrans[2];
  }
}

void CWalkingCameraControl::StrafeRight(float fAmount)
{
  // Calculate the new position
  float fTrans[3];
  fTrans[0] = fAmount * (COS(_an32RotXY[1]));
  fTrans[1] = 0;
  fTrans[2] = -fAmount * (SIN(_an32RotXY[1]));

  // Calculate collision
  bool bOK;
  if(_pfnCallback) 
    bOK = _pfnCallback(EWCC_Step, this, _afPosXYZ, fTrans, _pvUserCBData);
  else
    bOK = CameraCallback(EWCC_Step, _afPosXYZ, fTrans, _pvUserCBData);

  // Update only if allowed by the user callback function
  if(bOK) {
    _afPosXYZ[0] += fTrans[0];
    _afPosXYZ[2] += fTrans[2];
  }
}

void CWalkingCameraControl::ClimbUp(float fAmount)
{
  // Calculate the new position
  float fTrans[3];
  fTrans[0] = 0;
  fTrans[1] = fAmount;
  fTrans[2] = 0;

  // Calculate collision
  bool bOK;
  if(_pfnCallback) 
    bOK = _pfnCallback(EWCC_Climb, this, _afPosXYZ, fTrans, _pvUserCBData);
  else
    bOK = CameraCallback(EWCC_Climb, _afPosXYZ, fTrans, _pvUserCBData);

  // Update only if allowed by the user callback function
  if(bOK) _afPosXYZ[1] += fTrans[1];
}

void CWalkingCameraControl::ClimbDown(float fAmount)
{
  // Calculate the new position
  float fTrans[3];
  fTrans[0] = 0;
  fTrans[1] = -fAmount;
  fTrans[2] = 0;

  // Calculate collision
  bool bOK;
  if(_pfnCallback) 
    bOK = _pfnCallback(EWCC_Climb, this, _afPosXYZ, fTrans, _pvUserCBData);
  else
    bOK = CameraCallback(EWCC_Climb, _afPosXYZ, fTrans, _pvUserCBData);

  // Update only if allowed by the user callback function
  if(bOK) _afPosXYZ[1] += fTrans[1];
}

void CWalkingCameraControl::_NewSkybox_()
{
  if(_pcsSkybox) return;

  _pcsSkybox = _nothrow_new CSkybox(*this);
  if(!_pcsSkybox)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't create a new CWalkingCameraControl::Skybox instance");
}

bool CWalkingCameraControl::CameraCallback(EWCCContext eContext, const float afCurPosXYZ[3], float afTransFactorXYZ[3], void *pvUserData)
{
  UNUSED_PARAMS_4(eContext, afCurPosXYZ, afTransFactorXYZ, pvUserData);
  return(true);
}


//
// Skybox class for CWalkingCameraControl class
//
const float    CWalkingCameraControl::CSkybox::DefaultYtSensitivity  = 0.1f;
const float    CWalkingCameraControl::CSkybox::DefaultXrSensitivity  = 0.2f;
const uint32_t CWalkingCameraControl::CSkybox::DefaultViewAngleWidth = 1024;

CWalkingCameraControl::CSkybox::CSkybox(CWalkingCameraControl &camera)
: _camera(camera),            _sourceImage(NULL),         _yOffset(0.0f),
  _sYt(DefaultYtSensitivity), _sXr(DefaultXrSensitivity), _sYr(0.0f)
{
}

void CWalkingCameraControl::CSkybox::SetImageSource(CImage &sourceImage, C3DEnv *target3DEnv)
{
  _target = (target3DEnv != NULL) ? target3DEnv : C3DEnv::Active3DEnv();

  uint32_t Wr = _target->_tRB->un32Width;
  uint32_t Wi = sourceImage.GetWidth();

  uint32_t Wmax = max(Wi, Wr);
  uint32_t Wmin = min(Wi, Wr);
  uint32_t r    = (Wi <= Wr) ? (xprv::_DEGREECOUNT_ / DefaultViewAngleWidth) : (Wmax / Wmin);

  _sYr = float(Wr) / float(xprv::_DEGREECOUNT_) * float(r);

  if(r*Wmin != Wmax) 
    VError(VERR_InvalidParameter, _NFL_, "The source image width must at least meet one of these requirements :\n" \
                                         "  1. The image width is exactly the same as the render buffer width\n"   \
                                         "  2. The image width is fully divisible by the render buffer width\n"    \
                                         "  3. The render buffer width is fully divisible by the image width\n"    \
                                         "(current render buffer width = %d, source image width = %d)",
                                         Wr, Wi);
 
  _sourceImage = &sourceImage;
}

void CWalkingCameraControl::CSkybox::Render()
{
  if(C3DEnv::Active3DEnv() != _target)
    VError(VERR_NotSupported, _NFL_, "The active 3D environment is different with the 3D environment used\nin this CWalkingCameraControl::Skybox instance; skybox rendering is\nnot possible");

  // Get the source buffer parameters
  const uint32_t *sib = reinterpret_cast<const uint32_t*>(_sourceImage->GetBuffer());
        size_t    siw = _sourceImage->GetWidth();
        size_t    sih = _sourceImage->GetHeight();

  // Get the destination buffer parameters
  uint32_t *dib = reinterpret_cast<uint32_t*>(_target->_tRB->pun8Buffer);
  size_t    diw = _target->_tRB->un32Width;
  size_t    dih = _target->_tRB->un32Height;

  // Get the current camera's Y-axis rotation
  // and make sure that it is always positive
  ssize_t yRot = _camera.GetCurYRot() % xprv::_DEGREECOUNT_;
  if(yRot < 0) yRot += xprv::_DEGREECOUNT_;

  // Calculate the image starting X pixel
  size_t startX = size_t(yRot * _sYr) % siw;

  // Get the current camera's X-axis rotation
  // and make sure that it is always positive
  ssize_t xRot = _camera.GetCurXRot() % xprv::_DEGREECOUNT_;
  if(xRot < 0) xRot += xprv::_DEGREECOUNT_;

  // Get the current camera's Y-axis position
  // and make sure that it is always positive
  float yPos = (_camera.GetCurYPos() - _yOffset) * _sYt;

  // Calculate the image starting Y pixel
  ssize_t startY = ssize_t((xRot * _sYr) - yPos);
  while(startY < 0) startY += sih;
  while(startY >= ssize_t(sih)) startY -= sih;

  // Render the background
  size_t ssy = startY;
  for(size_t diy = 0; diy < dih; ++diy) {
    // Calculate the starting buffer pointer of the source image
    const uint32_t *stb = sib + ssy*siw;
    // Update the source buffer Y location
    ssy++; if(ssy >= sih) ssy = 0;
    // Render the horizontal scanline
    size_t ssx = startX;
    for(size_t dix = 0; dix < diw; ++dix) {
      // Copy the pixel value
      *dib++ = stb[ssx];
      // Update the source buffer X location
      ssx++; if(ssx >= siw) ssx = 0;
    }
  }
}

} // namespace vge
