/** \file CameraControl.h */
// ############################################################################################
//  CameraControl.h
//  Definition & implementation of CWalkingCameraControl class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___CAMERA_CONTROL_H_
#define _3D_ENVIRONMENT___CAMERA_CONTROL_H_

// VortexGE headers
#include "../2DManipulation/ImageClasses.h"
#include "Matrix3D.h"


namespace vge {

  // Foward reference from "lib/3DEnvironment/3DEnvironment.h"
  class C3DEnv;

  // Foward reference from within this file
  class CWalkingCameraControl;

  //!
  //! \brief Callback context
  //!
  //! \api beta
  enum EWCCContext { EWCC_Step, EWCC_Climb };

  //!
  //! \brief User callback function for CWalkingCameraControl
  //!
  //! \api beta
  typedef bool(*PFNVortexWCCCallback)(EWCCContext eContext, CWalkingCameraControl *pcsWCC, const float afCurPosXYZ[3], float afTransFactorXYZ[3], void *pvUserData);


  //!
  //! \brief The walking camera class
  //!
  //! \api pretty stable
  class VGE_EXPORT CWalkingCameraControl : VCLASS_UNCOPYABLE {
    public:
      class CSkybox;

    public:
      //!
      //! Construct the CWalkingCameraControl instance and set the user callback
      //! function and user data.
      //! \n\n
      //! If pfnWCCCalback is NULL, the CameraCallback() virtual member function
      //! will be called, otherwise the given callback function will be called.
      CWalkingCameraControl(PFNVortexWCCCallback pfnWCCCalback = NULL, void *pvUserData = NULL);
      //!
      //! Standard dtor.
      virtual ~CWalkingCameraControl() throw();

      // Initialization functions
      //!
      //! (Re-)set the camera user callback function.
      //! \n\n
      //! If pfnWCCCalback is NULL, the CameraCallback() virtual member function
      //! will be called, otherwise the given callback function will be called.
      void SetCallback(PFNVortexWCCCallback pfnWCCCalback, void *pvUserData);

      // Information functions
      //!
      //! Returns the current X position of the camera in the 3D world.
      inline float GetCurXPos() const { return(_afPosXYZ[0]); }
      //!
      //! Returns the current Y position of the camera in the 3D world.
      inline float GetCurYPos() const { return(_afPosXYZ[1]); }
      //!
      //! Returns the current Z position of the camera in the 3D world.
      inline float GetCurZPos() const { return(_afPosXYZ[2]); }
      //!
      //! Returns the current X rotation of the camera in the 3D world.
      inline int32_t GetCurXRot() const { return(_an32RotXY[0]); }
      //!
      //! Returns the current Y rotation of the camera in the 3D world.
      inline int32_t GetCurYRot() const { return(_an32RotXY[1]); }

      // Noncumulative setting functions
      //!
      //! Reset the camera's position and rotation, after the call
      //! the camera will be on (0, 0, 0) and facing to the north.
      inline void ResetAll();
      //!
      //! Set the camera's position.
      inline void SetPosition(float fX, float fY, float fZ);
      //!
      //! Set the camera's position.
      inline void SetPosition(const float afNewPosXYZ[3]);
      //!
      //! Set the camera's Y rotation.
      inline void SetYRotation(int32_t n32Amount);
      //!
      //! Make the camera looks to the up direction.
      inline void LookUp(int32_t n32Amount) { _an32RotXY[0] = n32Amount; }
      //!
      //! Make the camera looks to the straight (normal) direction.
      inline void LookStraight() { _an32RotXY[0] = 0; }
      //!
      //! Make the camera looks to the down direction.
      inline void LookDown(int32_t n32Amount) { _an32RotXY[0] = -n32Amount; }
      //!
      //! Rotate the camera to make it facing north.
      inline void RotateToTheNorth() { _an32RotXY[1] = 0; }
      //!
      //! Rotate the camera to make it facing north east.
      inline void RotateToTheNorthEast() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 45 / 360; }
      //!
      //! Rotate the camera to make it facing east.
      inline void RotateToTheEast() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 90 / 360; }
      //!
      //! Rotate the camera to make it facing south east.
      inline void RotateToTheSouthEast() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 135 / 360; }
      //!
      //! Rotate the camera to make it facing south.
      inline void RotateToTheSouth() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 180 / 360; }
      //!
      //! Rotate the camera to make it facing south west.
      inline void RotateToTheSouthWest() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 225 / 360; }
      //!
      //! Rotate the camera to make it facing west.
      inline void RotateToTheWest() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 270 / 360; }
      //!
      //! Rotate the camera to make it facing north west.
      inline void RotateToTheNorthWest() { _an32RotXY[1] = xprv::_DEGREECOUNT_ * 315 / 360; }

      // Cumulative updating functions
      //!
      //! Step backward as many as the given amount.
      void StepForward(float fAmount);
      //!
      //! Step foward as many as the given amount.
      void StepBackward(float fAmount);
      //!
      //! Strafe left as many as the given amount.
      void StrafeLeft(float fAmount);
      //!
      //! Strafe right as many as the given amount.
      void StrafeRight(float fAmount);
      //!
      //! Climb up as many as the given amount.
      void ClimbUp(float fAmount);
      //!
      //! Climb down as many as the given amount.
      void ClimbDown(float fAmount);
      //!
      //! Rotate left as many as the given amount.
      inline void RotateLeft(int32_t n32Amount);
      //!
      //! Rotate right as many as the given amount.
      inline void RotateRight(int32_t n32Amount);

      // Matrix functions
      //!
      //! Update the camera's matrix to reflect the current state of the camera.
      inline void UpdateCameraMatrix();
      //!
      //! Get the camera's matrix.
      inline CMatrix3D& GetCameraMatrix() { return(_csMatCamera); }
      //!
      //! Get the camera's matrix.
      inline const CMatrix3D& GetCameraMatrix() const { return(_csMatCamera); }
      //!
      //! Get the camera's Y matrix.
      inline CMatrix3D& GetYCameraMatrix() { return(_csMatRotY); }
      //!
      //! Get the camera's Y matrix.
      inline const CMatrix3D& GetYCameraMatrix()const { return(_csMatRotY); }

      // Access to the CSkybox class instance
      //!
      //! Direct access to the camera's CSkybox instance.
      inline CSkybox& iSkybox();
      //!
      //! Direct access to the camera's CSkybox instance.
      inline const CSkybox& iSkybox() const;

      // Camera callback function
      //!
      //! This virtual member function will be called if on initialization,
      //! either via ctor or via Initialize(), the pfnWCCCalback parameter
      //! is set to NULL, otherwise the given callback function will be called.
      virtual bool CameraCallback(EWCCContext eContext, const float afCurPosXYZ[3], float afTransFactorXYZ[3], void *pvUserData);

    protected:
      // Helper functions
      void _NewSkybox_();

      // Data
      PFNVortexWCCCallback _pfnCallback;  // Pointer to the user callback function
      void                *_pvUserCBData; // User callback data to be passed to the callback funtion
      CMatrix3D       _sse _csMatCamera;  // The camera matrix
      CMatrix3D       _sse _csMatRotX;    // The X rotation matrix
      CMatrix3D       _sse _csMatRotY;    // The Y rotation matrix
      float                _afPosXYZ[3];  // The camera position (in absolute local coordinate)
      int32_t              _an32RotXY[2]; // The camera position
      CSkybox             *_pcsSkybox;    // Distant scenery control (iSkybox)
  };

  void CWalkingCameraControl::UpdateCameraMatrix()
  {
    // Position
    _csMatCamera.SetToTranslationMatrix(-_afPosXYZ[0], -_afPosXYZ[1], -_afPosXYZ[2]);
    // Y rotation
    _csMatRotY._aafMatrix[0][0] = COS(-_an32RotXY[1]);
    _csMatRotY._aafMatrix[0][2] = -SIN(-_an32RotXY[1]);
    _csMatRotY._aafMatrix[2][0] = SIN(-_an32RotXY[1]);
    _csMatRotY._aafMatrix[2][2] = COS(-_an32RotXY[1]);
    _csMatCamera.MultMatrixCCN(_csMatRotY);
    // X rotation
    if(_an32RotXY[0]) {
      _csMatRotX._aafMatrix[1][1] =  COS(_an32RotXY[0]);
      _csMatRotX._aafMatrix[1][2] =  SIN(_an32RotXY[0]);
      _csMatRotX._aafMatrix[2][1] = -SIN(_an32RotXY[0]);
      _csMatRotX._aafMatrix[2][2] =  COS(_an32RotXY[0]);
      _csMatCamera.MultMatrixCCN(_csMatRotX);
    }
  }

  void CWalkingCameraControl::ResetAll()
  {
    _csMatCamera.SetToIdentityMatrix();
    _csMatRotX.SetToIdentityMatrix();
    _csMatRotY.SetToIdentityMatrix();
    _afPosXYZ[0] = _afPosXYZ[1] = _afPosXYZ[2] = 0.0f;
    _an32RotXY[0] = _an32RotXY[1] = 0;
  }

  void CWalkingCameraControl::SetPosition(float fX, float fY, float fZ)
  {
    _afPosXYZ[0] = fX;
    _afPosXYZ[1] = fY;
    _afPosXYZ[2] = fZ;
  }

  void CWalkingCameraControl::SetPosition(const float afNewPosXYZ[3])
  {
    _afPosXYZ[0] = afNewPosXYZ[0];
    _afPosXYZ[1] = afNewPosXYZ[1];
    _afPosXYZ[2] = afNewPosXYZ[2];
  }

  void CWalkingCameraControl::SetYRotation(int32_t n32Amount)
  {
    _an32RotXY[1] = n32Amount;
    _an32RotXY[1] &= (xprv::_DEGREECOUNT_ - 1);
  }

  void CWalkingCameraControl::RotateLeft(int32_t n32Amount)
  {
    _an32RotXY[1] -= n32Amount;
    _an32RotXY[1] &= (xprv::_DEGREECOUNT_ - 1);
  }

  void CWalkingCameraControl::RotateRight(int32_t n32Amount)
  {
    _an32RotXY[1] += n32Amount;
    _an32RotXY[1] &= (xprv::_DEGREECOUNT_ - 1);
  }

  CWalkingCameraControl::CSkybox& CWalkingCameraControl::iSkybox()
  {
    _if_unlikely(!_pcsSkybox) _NewSkybox_();
    return(*_pcsSkybox);
  }


  //!
  //! \brief Class used for rendering skybox for walking entities (eg. human)
  //!        in 1st person view, <b>can't</b> be used for flying entities 
  //!        (eg. airship)
  //!
  //! \api beta
  class VGE_EXPORT CWalkingCameraControl::CSkybox: VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct the CSkybox instance and attach it to the
      //! given CWalkingCameraControl instance.
      explicit CSkybox(CWalkingCameraControl &camera);

      // Initialization functions
      //!
      //! Use this to set the image to be displayed as the skybox.
      //! \param[in] sourceImage Image to be displayed, can either CImage or CAnimatedImage
      //! \param[in] target3DEnv 3D environment to be used, if NULL then the active one will be used.
      void SetImageSource(CImage &sourceImage, C3DEnv *target3DEnv = NULL);
      //!
      //! Use this to specify the starting Y coordinate (in 3D unit) to be assumed
      //! as the ground level (eg. human eyes has some distance with his feet, the
      //! Y offset here is the distance between the eyes and the feet).
      inline void SetCameraYOffset(float offset)
      { _yOffset = offset; }
    
      // Sensitivity setters
      //!
      //! Use this to set how much the scenerey image will be shifted in the
      //! vertical (Y) direction in case of the camera move as much as 1.0
      //! 3D unit in the world in the Y-axis direction (3D Y-axis translation).
      inline void SetYtSensitivity(float sensitivity = DefaultYtSensitivity)
      { _sYt = sensitivity; }
      //!
      //! Use this to set how much the scenerey image will be shifted in the
      //! vertical (Y) direction in case of the camera rotate as much as 1.0
      //! 3D unit in the world in the X-axis direction (3D X-axis rottation).
      inline void SetXrSensitivity(float sensitivity = DefaultXrSensitivity)
      { _sXr = sensitivity; }

      // Rendering functions
      //!
      //! Use this to render the Skybox
      void Render();

    protected:
      // Constants
      static const float    DefaultYtSensitivity;
      static const float    DefaultXrSensitivity;
      static const uint32_t DefaultViewAngleWidth;

      // Data
      CWalkingCameraControl &_camera;      // Owning camera
      C3DEnv                *_target;      // Target 3D environment
      CImage                *_sourceImage; // Source image to be used
      float                 _yOffset;      // Camera's Y offset
      float                 _sYt, _sXr;    // Sensitivity to the camera's Y translation & X rotation
      float                 _sYr;          // Sensitivity to the camera's Y rotation (will be calculated)
  };

} // namespace vge
#endif
