/** \file Miscellaneous3DStructs.h */
// ############################################################################################
//  Miscellaneous3DStructs.h
//  Definition and implementation of TFloatXY, _TPoint2D, _TVector3D and
//  TTexUV structures and their related convenience functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___MISCELLANEOUS_3D_STRUCTS_H_
#define _3D_ENVIRONMENT___MISCELLANEOUS_3D_STRUCTS_H_

// VortexGE headers
#include "../Base/SIMDTypes.h"


namespace vge {

  //!
  //! \brief Doublet float XY structure
  //!
  //! \api pretty stable
  struct TFloatXY {
    float x, y;
  
    inline TFloatXY(float x_ = 0, float y_ = 0)
    : x(x_), y(y_)
    {}

    inline TFloatXY &operator+=(const TFloatXY &val)
    { x += val.x; y += val.y; return(*this); }
  
    inline TFloatXY &operator-=(const TFloatXY &val)
    { x -= val.x; y -= val.y; return(*this); }

    inline TFloatXY operator+(const TFloatXY &val)
    { TFloatXY tmp; tmp.x = x + val.x; tmp.y = y + val.y; return(tmp); }

    inline TFloatXY operator-(const TFloatXY &val)
    { TFloatXY tmp; tmp.x = x - val.x; tmp.y = y - val.y; return(tmp); }
  };


  //! \brief Triplet int32_t XYZ structure
  //!
  //! \api pretty stable
  struct TInt32XYZ {
    int32_t x, y, z;
  
    inline TInt32XYZ(int32_t x_ = 0, int32_t y_ = 0, int32_t z_ = 0)
    : x(x_), y(y_), z(z_)
    {}

    inline TInt32XYZ &operator+=(const TInt32XYZ &val)
    { x += val.x; y += val.y; z += val.z; return(*this); }
  
    inline TInt32XYZ &operator-=(const TInt32XYZ &val)
    { x -= val.x; y -= val.y; z -= val.z; return(*this); }

    inline TInt32XYZ operator+(const TInt32XYZ &val)
    { TInt32XYZ tmp; tmp.x = x + val.x; tmp.y = y + val.y; tmp.z = z + val.z; return(tmp); }

    inline TInt32XYZ operator-(const TInt32XYZ &val)
    { TInt32XYZ tmp; tmp.x = x - val.x; tmp.y = y - val.y; tmp.z = z - val.z; return(tmp); }
  };


  //!
  //! \internal
  //! \brief 2D point structure used when rendering
  //!
  //! \api pretty stable
  struct _sse _TPoint2D {
#ifndef __USE_AVEC__
    int32_t n32X, n32Y;       // Screen (X, Y) coordinate of this face
    int32_t n32Z;             // 1/Z value of this face
    int32_t n32F;             // Fog intensity of this face
    int32_t n32U, n32V;       // Texture coordinates of this face
    int32_t n32R, n32G, n32B; // Color of this face
#else
    // NOTE: Do not change the order !!!
    int32_t _sse n32X;
    int32_t      n32Y, ___dummy1, ___dummy2;
    int32_t      n32Z, n32F, n32U, n32V;
    int32_t      n32R, n32G, n32B, ___dummy3;
#endif
  };

  inline bool operator==(_TPoint2D &t1, _TPoint2D &t2)
  { return((t1.n32X == t2.n32X) && (t1.n32Y == t2.n32Y)); }


  //!
  //! \internal
  //! \brief 3D vector structure
  //!
  //! \api pretty stable
  struct _TVector3D {
    float fX, fY, fZ;    // The XYZ direction vector
#ifdef _ALIGN_128_BITS_
    float _fdummy_to_make_the_XYZ_vector_128_bits_long_;
#endif
    float fXt, fYt, fZt; // The transformed XYZ direction vector
#ifdef _ALIGN_128_BITS_
    float _fdummy_to_make_the_XtYtZt_vector_128_bits_long_;
#endif
  };


  //! 
  //! \brief The texture UV structure
  //!
  //! \api pretty stable
  struct TTexUV {
    float fU, fV;
  };


  // Internal inline functions implementation related to
  // Sutherland-Hodgman algorithm
  namespace xprv {

    inline void _SHCopy(TTexUV &tClip, const TTexUV &t1)
    {
      tClip.fU = t1.fU;
      tClip.fV = t1.fV;
    }

    inline void _SHClip(TTexUV &tClip, const TTexUV &t1, const TTexUV &t2, float fT)
    {
      tClip.fU = t1.fU + ((t2.fU - t1.fU) * fT);
      tClip.fV = t1.fV + ((t2.fV - t1.fV) * fT);
    }

  } // namespace xprv

} // namespace vge
#endif
