/** \file Particle2D.h */
// ############################################################################################
//  Particle2D.h
//  Definition & implementation of CSnow2D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___PARTICLE_2D_H_
#define _2D_MANIPULATION___PARTICLE_2D_H_

// VortexGE headers
#include "../Base/Random.h"
#include "../Base/SpecialClasses.h"


namespace vge {

  // Foward reference from "lib/3DEnvironment/3DEnvironment.h"
  class C3DEnv;


  //!
  //! \brief Particle class used for simulating snow
  //!
  //! \api beta
  class VGE_EXPORT CSnow2D : VCLASS_UNCOPYABLE {
    public:
      CSnow2D();
      ~CSnow2D() throw();

      // Direct access to the particles' random number generator
      //!
      //! Direct access to the particles' random number generator
      inline CRandomCoreUInt32 &RandomNumberGenerator()
      { return(_rnGen); }
      //!
      //! Direct access to the particles' random number generator
      inline const CRandomCoreUInt32 &RandomNumberGenerator() const
      { return(_rnGen); }

      // Initialization functions
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeAll();
      //!
      //! Initialize the class and set the snow parameters.
      //! \param[in] pcsV3DEnv         3D enviroment to be used when rendering the particles.
      //! \param[in] un32TotalParticle Total particles to be genereted (displayed).
      //! \param[in] fMaxSimZ          Maximum simulated Z buffer value.
      //! \param[in] n32XMin           X coordinate on which particles will start appearing.
      //! \param[in] n32XMax           X coordinate on which particles will disappear.
      //! \param[in] n32YStartMin      Minimum Y coordinate on which particles will start appearing.
      //! \param[in] n32YStartMax      Maximum Y coordinate on which particles will start appearing.
      //! \param[in] n32YEndMin        Minimum Y coordinate on which particles will disappear.
      //! \param[in] n32YEndMax        Maximum Y coordinate on which particles will disappear.
      //! \param[in] fXSpeed           X-axis movement speed of the particles.
      //! \param[in] fYSpeed           Y-axis movement speed of the particles.
      //! \param[in] un8R,un8G,un8B    Color of the particles.
      void Initialize(C3DEnv *pcsV3DEnv, uint32_t un32TotalParticle,
                      float fMaxSimZ, int32_t n32XMin, int32_t n32XMax,
                      int32_t n32YStartMin, int32_t n32YStartMax,
                      int32_t n32YEndMin, int32_t n32YEndMax,
                      float fXSpeed, float fYSpeed,
                      uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! (Re-)set the snow parameters.
      //! See parameters explanation for the Initialize() function.
      void SetAllSpec(float fMaxSimZ, int32_t n32XMin, int32_t n32XMax,
                      int32_t n32YStartMin, int32_t n32YStartMax,
                      int32_t n32YEndMin, int32_t n32YEndMax,
                      float fXSpeed, float fYSpeed,
                      uint8_t un8R, uint8_t un8G, uint8_t un8B);

      // Updating functions
      //!
      //! Update the particles appearance (move to the next frame).
      void Update();
      //!
      //! (Re-)set the particles X-axis movement speed.
      //! \param[in] fXSpeed X-axis movement speed of the particles.
      inline void SetXSpeed(float fXSpeed)
      { _tSpec.fXSpeed = fXSpeed; }
      //!
      //! (Re-)set the particles color.
      //! \param[in] un8R,un8G,un8B Color of the particles.
      inline void SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B);

      // Rendering functions
      //!
      //! Render the particles in solid mode, without using Z-buffer
      //! (must be the last function called in a rendering sequence).
      void SRenderNZ();
      //!
      //! Render the particles in solid, using Z-buffer.
      void SRenderWZ();
      //!
      //! Render the particles in fogged & solid mode, using Z-buffer.
      void FogSRenderWZ();
      //!
      //! Render the particles in additive mode, without using Z-buffer
      //! (must be the last function called in a rendering sequence).
      void ARenderNZ();
      //!
      //! Render the particles in averaging mode, without using Z-buffer
      //! (must be the last function called in a rendering sequence).
      void VRenderNZ();

  protected:
      struct _R2Spec {
        uint32_t un32TotPar;                   // Total particles created
        float    fMaxSimZ;                     // Maximum Z value (for simulating Z distance)
        int32_t  n32XMin, n32XMax;             // X coordinates limit
        int32_t  n32YStartMin, n32YStartMax;   // Y start coordinates limit
        int32_t  n32YEndMin, n32YEndMax;       // Y end coordinates limit
        float    fXSpeed, fYSpeed;             // XY Movement speed
        int32_t  n32XRange;                    // Calculated X coordinate range
        int32_t  n32YStartRange, n32YEndRange; // Calculated Y coordinate ranges
        uint8_t  un8R, un8G, un8B;             // Particles' Color
      };
      struct _R2Data {
        bool    bValid;   // Validity flag
        float   fX, fY;   // Current position
        int32_t n32YMax;  // Maximum Y position
        float   fZ;       // The simulated Z distance
        int32_t n32ZBuff; // The simulated Z-buffer value
      };

      // Data
      C3DEnv            *_pcs3DEnv; // Pointer to a dedicated 3D environment
      _R2Spec            _tSpec;    // Specification data of the particles
      _R2Data           *_ptPar;    // Array of particles
      CRandomCoreUInt32  _rnGen;    // Core random number generator
      CRandomFloat       _rnFloat;  // Floating points random number generator
  };

  void CSnow2D::SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
  {
    _tSpec.un8R  = un8R;
    _tSpec.un8G  = un8G;
    _tSpec.un8B  = un8B;
  }

} // namespace vge
#endif
