// ############################################################################################
//  AOSSEnvironment.cc
//  Implementation of COssAudioEnv class using OSS audio system
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#include "../../VConfig.h"
#ifdef __USE_OSS_AUDIO__

// VortexGE headers
#include "../MidiPlayer.h"
#include "AOSSSoundBuffer.h"

// OSS headers
#include <sys/soundcard.h>


namespace vge {

//
// The OSS audio environment implementation
//
VCLASS_SINGLETON_IMPLEMENT(COssAudioEnv)

COssAudioEnv *COssAudioEnv::_instance = NULL;

COssAudioEnv::COssAudioEnv()
{
  if(!COssAudioEnv::_instance) {
    CAudioEnvCImp::_instAudioEnv = this;
    COssAudioEnv::_instance = this;
  }

  CAudioEnvCImp::_tCD = &_tD;

  memset(_atPSB, 0, sizeof(_atPSB[0]) * _MAX_SB_);
  memset(&_tD, 0, sizeof(_tD));
  _tD.n32DSPFD = -1;
}

COssAudioEnv::~COssAudioEnv() throw()
{
  // Terminate threads
  try {
    _csTWriter.Stop();
  }
  catch(...) {
    _csTWriter.Kill();
  }

  // Close audio device
  if(_tD.n32DSPFD >= 0) {
    int dummy;
    ioctl(_tD.n32DSPFD, SNDCTL_DSP_RESET, &dummy);
    close(_tD.n32DSPFD);
  }

  // Free memory
  if(_tD.pucPBuff) free(_tD.pucPBuff);

  // Set reference to NULL
  if(COssAudioEnv::_instance == this) COssAudioEnv::_instance = NULL;
}

void COssAudioEnv::DoQuickInit(const char *pcDeviceName)
{
  if(pcDeviceName)
    OpenDevice(pcDeviceName);
  else {
    char acBuff[PATH_MAX];
    QueryDeviceName(acBuff, sizeof(acBuff));
    OpenDevice(acBuff);
  }
  InitializeDevice();
  ActivateDevice();
}

// NOTE: 1. If pcResult is NULL, this function will only check the
//          availability of the device.
//       2. In case of the system has multiple devices, only the 1st
//          device name will be returned. If the first device can't
//          be opened, OpenDevice() will adjust the device name to
//          open the 2nd, 3rd, and so.
size_t COssAudioEnv::QueryDeviceName(char *pcResult, size_t bufferSize)
{
  // For processing
  struct stat tStat;
  char        acTmp[PATH_MAX];
  string      ossName;

  // Empty the destination string
  if(pcResult) pcResult[0] = 0;

  // Search for /dev/dsp*
  if((stat("/dev/dsp", &tStat) == 0) && S_ISCHR(tStat.st_mode))
    ossName = "/dev/dsp";
  else {
    for(size_t i = 0; i < 9; ++i) {
      snprintf(acTmp, sizeof(acTmp), "/dev/dsp%zd", i);
      if((stat(acTmp, &tStat) == 0) && S_ISCHR(tStat.st_mode)) {
        ossName = acTmp;
        break;
      }
    }
  }

  // Search for /dev/sound/dsp*
  if(ossName.empty() && (stat("/dev/sound", &tStat) == 0) && S_ISDIR(tStat.st_mode)) {
    if((stat("/dev/sound/dsp", &tStat) == 0) && S_ISCHR(tStat.st_mode))
      ossName = "/dev/sound/dsp";
    else {
      for(size_t i = 0; i < 9; ++i) {
        snprintf(acTmp, sizeof(acTmp), "/dev/sound/dsp%zd", i);
        if((stat(acTmp, &tStat) == 0) && S_ISCHR(tStat.st_mode)) {
          ossName = acTmp;
          break;
        }
      }
    }
  }

  // Search for /dev/audio*
  if(ossName.empty()) {
    if((stat("/dev/audio", &tStat) == 0) && S_ISCHR(tStat.st_mode))
      ossName = "/dev/audio";
    else {
      for(size_t i = 0; i < 9; ++i) {
        snprintf(acTmp, sizeof(acTmp), "/dev/audio%zd", i);
        if((stat(acTmp, &tStat) == 0) && S_ISCHR(tStat.st_mode)) {
          ossName = acTmp;
          break;
        }
      }
    }
  }

  // Check if we got something
  if(!ossName.empty()) {
    if(pcResult && bufferSize) safe_strcpy(pcResult, ossName.c_str(), bufferSize);
    return(ossName.length() + 1);
  }

  // Not found
  VError(VERR_NotFound, _NFL_, "No usable OSS audio device could be found");
}

bool COssAudioEnv::_AdjustDeviceName_(char *pcDeviceName)
{
  uint32_t  len = strlen(pcDeviceName);
  char     *num = &pcDeviceName[len - 1];
  switch(*num) {
    case '0': *num = '1'; break;
    case '1': *num = '2'; break;
    case '2': *num = '3'; break;
    case '3': *num = '4'; break;
    case '4': *num = '5'; break;
    case '5': *num = '6'; break;
    case '6': *num = '7'; break;
    case '7': *num = '8'; break;
    case '8': *num = '9'; break;
    case '9': return(false);
    default : pcDeviceName[len]     = '0';
              pcDeviceName[len + 1] = 0;
  };
  return(true);
}

void COssAudioEnv::OpenDevice(const char *pcDeviceName)
{
  // Check
  if(_tD.n32DSPFD > 0) return;

  // If the user have a preferred device name set in the environment variable,
  // try that first
  char *pcEnv = getenv("VGE_AUDIO_DEVICE");
  if(pcEnv) {
    // Open the device
    _tD.n32DSPFD = open(pcEnv, O_WRONLY);
    if(_tD.n32DSPFD >= 0) {
      // Copy the device name
      safe_strcpy(_tD.acDevName, pcEnv, sizeof(_tD.acDevName));
      // Reset the device and finish
      int dummy;
      ioctl(_tD.n32DSPFD, SNDCTL_DSP_RESET, &dummy);
      return;
    }
  }

  // Try to open the device given in the pcDeviceName parameter
  char acDevName[PATH_MAX];
  safe_strcpy(acDevName, pcDeviceName, sizeof(acDevName));
  _tD.n32DSPFD = open(acDevName, O_WRONLY);

  // If can't be opened, try to open the next device
  if(_tD.n32DSPFD < 0) {
    // Adjust the device name and try to open it
    while(_AdjustDeviceName_(acDevName)) {
      _tD.n32DSPFD = open(acDevName, O_WRONLY);
      if(_tD.n32DSPFD > 0) break;
    }
    // No device can be opened ?
    if(_tD.n32DSPFD < 0)
      VError(VERR_OpenDevice, _NFL_, "Couldn't open audio device \"%s\"", pcDeviceName);
  }

  // Copy the device name
  safe_strcpy(_tD.acDevName, acDevName, sizeof(_tD.acDevName));

  // Reset the audio device
  int dummy;
  ioctl(_tD.n32DSPFD, SNDCTL_DSP_RESET, &dummy);
}

static int32_t _VSBF2OSS_(ESoundBufferFormat eFormat)
{
  switch(eFormat) {
    case ESB_Undefined   : return(AFMT_S16_LE); // If not defined, we prefer 16-bit signed LE
    case ESB_Unknown     : return(AFMT_S16_LE); // If unknown, we prefer 16-bit signed LE
    case ESB_Signed8     : return(AFMT_S8);
    case ESB_Unsigned8   : return(AFMT_U8);
    case ESB_Signed16LE  : return(AFMT_S16_LE);
    case ESB_Unsigned16LE: return(AFMT_U16_LE);
    case ESB_Signed16BE  : return(AFMT_S16_BE);
    case ESB_Unsigned16BE: return(AFMT_U16_BE);
    default              : VError(VERR_InvalidParameter, _NFL_, "Invalid ESoundBufferFormat");
  }
}

void COssAudioEnv::InitializeDevice(const TAudioSettings *ptPreferred, TAudioSettings *ptRealized)
{
  // Check
  if(_tD.n32DSPFD < 0)
    VError(VERR_DeviceNotOpenend, _NFL_, "Audio device is not opened");

  // Preparations
  TAudioSettings *ptP, tP = { ESB_Undefined, 44100, 2 };
  TAudioSettings *ptR, tR;
  if(ptPreferred)
    ptP = const_cast<TAudioSettings*>(ptPreferred);
  else
    ptP = &tP;
  if(ptRealized)
    ptR = ptRealized;
  else
    ptR = &tR;

  // Get supported audio formats
  int ioctlTmp;
  if(ioctl(_tD.n32DSPFD, SNDCTL_DSP_GETFMTS, &ioctlTmp) < 0)
    VError(VERR_GetDeviceParam, _NFL_, "Couldn't get supported audio formats");

  // Find usable audio format
  if     (ioctlTmp & AFMT_S16_LE) ptR->eFormat = ESB_Signed16LE;
  else if(ioctlTmp & AFMT_U16_LE) ptR->eFormat = ESB_Unsigned16LE;
  else if(ioctlTmp & AFMT_S8)     ptR->eFormat = ESB_Signed8;
  else if(ioctlTmp & AFMT_U8)     ptR->eFormat = ESB_Unsigned8;
  else if(ioctlTmp & AFMT_S16_BE) ptR->eFormat = ESB_Signed16BE;
  else if(ioctlTmp & AFMT_U16_BE) ptR->eFormat = ESB_Unsigned16BE;
  else
    VError(VERR_NotSupported, _NFL_,
           "The audio device does not support any audio format compatible with this application");

  // Find usable channels
  ioctlTmp = clip_value<int>(1, ptP->n32Channels, 2);
  if(ioctlTmp != ptP->n32Channels)
    VWarning(_NFL_, "Only 1 channel or 2 channels audio is supported by COssAudioEnv");
  if(ioctl(_tD.n32DSPFD, SNDCTL_DSP_CHANNELS, &ioctlTmp) < 0)
    VError(VERR_SetDeviceParam, _NFL_, "Couldn't set audio channels");
  ptR->n32Channels = ioctlTmp;

  // Find usable frequency
  ioctlTmp = ptP->n32Frequency;
  if(ioctl(_tD.n32DSPFD, SOUND_PCM_WRITE_RATE, &ioctlTmp) < 0)
    VError(VERR_SetDeviceParam, _NFL_, "Couldn't set audio frequency (sampling rate)");
  ptR->n32Frequency = ioctlTmp;

  // Calculate buffer size
  //  NOTE: * The bigger the buffer the better it perfoms on slow computer,
  //          however, bigger buffer will cause more delay when playing sound.
  //        * Experimentally, 2048 is the most acceptable.
  const int32_t n32BuffSize = 2048;
  if((ptR->eFormat == ESB_Unsigned8) || (ptR->eFormat == ESB_Signed8))
    ioctlTmp = n32BuffSize * ptR->n32Channels;
  else
    ioctlTmp = n32BuffSize * ptR->n32Channels * 2;

  // Calculate fragment size
  int32_t n32Frag;
  for(n32Frag = 0; (0x01 << n32Frag) < ioctlTmp; n32Frag++);
  n32Frag |= 0x00020000;

  // Reopen audio device
  close(_tD.n32DSPFD);
  _tD.n32DSPFD = open(_tD.acDevName, O_WRONLY);
  if(_tD.n32DSPFD < 0)
    VError(VERR_OpenDevice, _NFL_, "Couldn't reopen audio device \"%s\"", _tD.acDevName);
  ioctl(_tD.n32DSPFD, SNDCTL_DSP_RESET, &ioctlTmp);

  // Set fixed DSP settings
  int format = _VSBF2OSS_(ptR->eFormat);
  if(ioctl(_tD.n32DSPFD, SNDCTL_DSP_SETFMT, &format) < 0)
    VError(VERR_SetDeviceParam, _NFL_, "Couldn't set audio format");
  ioctlTmp = n32Frag;
  if(ioctl(_tD.n32DSPFD, SNDCTL_DSP_SETFRAGMENT, &ioctlTmp) < 0)
    VError(VERR_SetDeviceParam, _NFL_, "Couldn't set audio fragment");
  ioctlTmp = ptR->n32Channels;
  if(ioctl(_tD.n32DSPFD, SNDCTL_DSP_CHANNELS, &ioctlTmp) < 0)
    VError(VERR_SetDeviceParam, _NFL_, "Couldn't set audio channels");
  ioctlTmp = ptR->n32Frequency;
  if(ioctl(_tD.n32DSPFD, SOUND_PCM_WRITE_RATE, &ioctlTmp) < 0)
    VError(VERR_SetDeviceParam, _NFL_, "Couldn't set audio frequency (sampling rate)");

  // Check and copy formats
  if(ptR->eFormat == ESB_Unsigned8) {
    _tD.tFormat.un8Bits   = 8;
    _tD.tFormat.un8Sign   = 0;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Signed8) {
    _tD.tFormat.un8Bits   = 8;
    _tD.tFormat.un8Sign   = 1;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Unsigned16LE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 0;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Unsigned16BE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 0;
    _tD.tFormat.un8Endian = 1;
  }
  else if(ptR->eFormat == ESB_Signed16LE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 1;
    _tD.tFormat.un8Endian = 0;
  }
  else if(ptR->eFormat == ESB_Signed16BE) {
    _tD.tFormat.un8Bits   = 16;
    _tD.tFormat.un8Sign   = 1;
    _tD.tFormat.un8Endian = 1;
  }
  else
    VError(VERR_NotSupported, _NFL_,
           "The audio device does not support any audio format compatible with this application");

  _tD.tFormat.un8Channels   = ptR->n32Channels;
  _tD.tFormat.un32Frequency = ptR->n32Frequency;
  _tD.un32FragSize  = 1 << (n32Frag & 0x0000FFFF);
  _tD.un32FragSize2 = _tD.un32FragSize >> 1;
  _tD.un32FragSize4 = _tD.un32FragSize >> 2;
  _tD.un32PBuffSize = _tD.un32FragSize * 4;
  _eFormat          = ptR->eFormat;

  // Allocate memory for primary buffer
  _tD.pucPBuff = reinterpret_cast<u_char*>(malloc(_tD.un32PBuffSize));
  if(_tD.pucPBuff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for primary sound buffer");
  _tD.pucPBuffRead  = _tD.pucPBuff;
  _tD.pucPBuffWrite = _tD.pucPBuff;
  _tD.pucPBuffMax   = _tD.pucPBuff + _tD.un32PBuffSize - _tD.un32FragSize;
  _tD.un8ToWrite    = 0;
  memset(_tD.pucPBuff, 0, _tD.un32PBuffSize);

  // Initialize threads
  _csTWriter.Initialize(COssAudioEnv::_ThreadProc, this);

  // Initialize sound mixing function
  // Assuming that this system is a little-endian (x86)
  //! \TODO Convert it to table !!!
  if(_tD.tFormat.un8Channels == 1) {
    if(_tD.tFormat.un8Bits == 8) {
      if(_tD.tFormat.un8Sign == 0)     _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU8m;   // U8 mono
      else                             _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS8m;   // S8 mono
    }
    else {
      if(_tD.tFormat.un8Sign == 0) {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU16m;  // U16 LE mono
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerU16m; // U16 BE mono
      }
      else {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS16m;  // S16 LE mono
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerS16m; // S16 BE mono
      }
    }
  }
  else {
    if(_tD.tFormat.un8Bits == 8) {
      if(_tD.tFormat.un8Sign == 0)     _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU8s;   // U8 stereo
      else                             _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS8s;   // S8 stereo
    }
    else {
      if(_tD.tFormat.un8Sign == 0) {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerU16s;  // U16 LE stereo
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerU16s; // U16 BE stereo
      }
      else {
        if(_tD.tFormat.un8Endian == 0) _pfnDoSoundMixing = CAudioEnvCImp::_DSPMixerS16s;  // S16 LE stereo
        else                           _pfnDoSoundMixing = CAudioEnvCImp::_IDSPMixerS16s; // S16 BE stereo
      }
    }
  }

  // Show info
  VInfo(_NFL_, "Realized audio device settings :\n  Device name         = %s\n  Format              = %s\n  Fragment size       = %u bytes\n  Channels            = %d\n  Frequency           = %d Hz\n  Primary buffer size = %u",
        _tD.acDevName, VSoundBufferFormatName(_eFormat),
        _tD.un32FragSize, _tD.tFormat.un8Channels,
        _tD.tFormat.un32Frequency, _tD.un32PBuffSize);
  if(ptPreferred) {
    if(ptR->eFormat!=ptP->eFormat || ptR->n32Channels!=ptP->n32Channels || ptR->n32Frequency!=ptP->n32Frequency)
      VInfo(_NFL_, "The realized audio settings are different than the preferred settings");
  }
}

void COssAudioEnv::ActivateDevice()
{
  // Check
  if(_tD.n32DSPFD < 0)
    VError(VERR_DeviceNotOpenend, _NFL_, "Audio device is not opened");

  // Clear buffer
  _ClearBuffer_(_tD.pucPBuff, _tD.un32PBuffSize, _tD.tFormat.un8Sign, _tD.tFormat.un8Bits);

  // Activate writer thread
  _csTWriter.Start();
}

void COssAudioEnv::DeactivateDevice()
{
  // Check
  if(_tD.n32DSPFD < 0) return;

  // Deactivate writer thread
  try {
    _csTWriter.Stop();
  }
  catch(...) {
    _csTWriter.Kill();
  }
}

void COssAudioEnv::GetAudioSpec(uint32_t *un32Channels, uint32_t *un32Bits, uint32_t *un32Rates, ESoundBufferFormat *eFormat, uint32_t *un32FragSize) const
{
  // Check
  if(_tD.n32DSPFD < 0) {
    *un32Channels = 0;
    *un32Bits     = 0;
    *un32Rates    = 0;
    *eFormat      = ESB_Unknown;
    *un32FragSize = 0;
  }

  // Return the format
  *un32Channels = _tD.tFormat.un8Channels;
  *un32Bits     = _tD.tFormat.un8Bits;
  *un32Rates    = _tD.tFormat.un32Frequency;
  *eFormat      = _eFormat;
  *un32FragSize = _tD.un32FragSize;
}

void COssAudioEnv::SetMidiPlayer(CMidiPlayer *pcsPlayer)
{
  _pcsMidiPlayer = pcsPlayer;
}

void COssAudioEnv::_ClearBuffer_(u_char *pucBuff, uint32_t un32BuffSize, uint8_t un8Sign, uint8_t un8Bits)
{
  if(un8Sign)
    memset(pucBuff, 0x00, un32BuffSize);
  else if(un8Bits == 8)
    memset(pucBuff, 0x80, un32BuffSize);
  else {
    uint16_t *buf16 = reinterpret_cast<uint16_t*>(pucBuff);
    for(size_t i = 0; i < (un32BuffSize >> 1); ++i) *buf16++ = 0x8000;
  }
}

CSoundBuffer *COssAudioEnv::NewSoundBuffer()
{
  return(_nothrow_new COssSoundBuffer);
}

uint64_t COssAudioEnv::_ThreadProc(TThreadData *ptThreadData)
{
  VInfo(_NFL_, "OSS thread activated");

  while(!(*ptThreadData->pbExit)) {
    // Suspended ?
    _if_unlikely(*ptThreadData->pbSuspend) {
      usleep(100000);
      continue;
    }
    usleep(1000);

    // * If there is no available sound buffer data, don't write
    //   to the audio device.
    // * Using this mechanism, the display can be synced with the
    //   audio if the sound buffer updating function is called
    //   from the display (frame updating) loop.
    if(COssAudioEnv::_instance->_tD.un8ToWrite > 3) continue;

    // Mix with the MIDI data
    if(COssAudioEnv::_instance->_pcsMidiPlayer) {
       COssAudioEnv::_instance->_pcsMidiPlayer->_MixWithBuffer(COssAudioEnv::_instance->_tD.pucPBuffRead,
                                                               COssAudioEnv::_instance->_tD.un32FragSize,
                                                               ptThreadData->pbExit);
    }

    // Write data
    // Original version : write(COssAudioEnv::_instance->_tD.n32DSPFD,
    //                          COssAudioEnv::_instance->_tD.pucPBuffRead,
    //                          COssAudioEnv::_instance->_tD.un32FragSize);
    //
    // Hack for kernel 2.6.x ALSA's OSS emulation :
    //
    uint32_t un32FS2 = COssAudioEnv::_instance->_tD.un32FragSize >> 1;
    write(COssAudioEnv::_instance->_tD.n32DSPFD, COssAudioEnv::_instance->_tD.pucPBuffRead, un32FS2);
    write(COssAudioEnv::_instance->_tD.n32DSPFD, COssAudioEnv::_instance->_tD.pucPBuffRead + un32FS2, un32FS2);

    // Increment the "ToWrite" value
    COssAudioEnv::_instance->_csMutex.Lock();
    COssAudioEnv::_instance->_tD.un8ToWrite++;
    COssAudioEnv::_instance->_csMutex.Unlock();

    // Increment the read pointer
    COssAudioEnv::_instance->_tD.pucPBuffRead += COssAudioEnv::_instance->_tD.un32FragSize;
    if(COssAudioEnv::_instance->_tD.pucPBuffRead > COssAudioEnv::_instance->_tD.pucPBuffMax) COssAudioEnv::_instance->_tD.pucPBuffRead = COssAudioEnv::_instance->_tD.pucPBuff;
  }

  VInfo(_NFL_, "OSS thread deactivated");
  return(0);
}

} // namespace vge
#endif
