/** \file LookUpTables.h */
// ############################################################################################
//  LookUpTables.h
//  Definition of various LUTs
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file is partialy adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___LOOK_UP_TABLES_H_
#define _BASE___LOOK_UP_TABLES_H_

// VortexGE headers
#include "VBasics.h"


namespace vge {

  // Internal LUTs and their support functions declaration 
  namespace xprv {

    // Sine and cosinde table size, must be in the power of two
    const uint32_t _DEGREECOUNT_ = uint32_t(4096);


    // LUT initialization functions
    //!
    //! \internal Initialize the sine and cosine table.
    extern void _VInitSinCosTable_();
    //!
    //! \internal Initialize the lighting table.
    extern void _VInitLightingTable_();


    // Trigonometry LUTS
    // These LUTs contains the sine and cosine values from 0.1 to 360.0 degree
    extern VGE_EXPORT float _afSinTable_[_DEGREECOUNT_];
    extern VGE_EXPORT float _afCosTable_[_DEGREECOUNT_];

    // Lighting LUTs
    // Each LUT represent R, G, and B components of this following lighting equation :
    //   LUT[(C1 << 5) | C2] = (C1 * C2) / 31
    // C1 and C2 are swapable.
    // If C1 is assumed as the panel's color then C2 is the light's color.
    // If C2 is assumed as the panel's color then C1 is the light's color.
    extern uint32_t _aun32RLightTable_[32 * 32];
    extern uint32_t _aun32GLightTable_[32 * 32];
    extern uint32_t _aun32BLightTable_[32 * 32];

    // Fogged-lighting LUTs
    // Each LUT represent R, G, and B components of this following fogged-lighting equation :
    //   LUT[(FI << 10) | (C1 << 5) | C2] = ((FI * FC) + ((31 - FI) * CT)) / 31;
    // with :
    //   CT = ((C1 * C2) / 31) & 0x1F
    // FI is the fog's intensity and FC is the fog's color for each R, G, B components.
    // C1 and C2 are the same as in the lighting LUTs.
    extern uint32_t *_pun32RFogLightTable_;
    extern uint32_t *_pun32GFogLightTable_;
    extern uint32_t *_pun32BFogLightTable_;

    // Fogged-color LUTs
    // Each LUT represent R, G, and B components of this following fogged-color equation :
    //   LUT[(FI << 5) | CL] = ((FI * FC) + ((31 - FI) * CL)) / 31
    // FI is the fog intensity and FC is the fog color for each R, G, B components.
    // CL can be the panel's color or the light's color.
    extern uint32_t *_pun32RFogColorTable_;
    extern uint32_t *_pun32GFogColorTable_;
    extern uint32_t *_pun32BFogColorTable_;

    // Color addition LUT
    // This LUT represent the result of this following addition equation :
    //   LUT[(C1 << 5) | C2] = C1 + C2
    // with restriction :
    //   0<= LUT[*] <= 31
    // C1 and C2 are swapable.
    extern uint8_t _aun8CAddTable_[32 * 32];

    // Transparency LUTs
    // This LUT below represent the result of this following mean equation :
    //   LUT[(C1 << 5) | C2] = ((C1 + C2) / 2) << 3
    // C1 and C2 are swapable.
    extern uint8_t _aun8CMeanTable_[32 * 32];
    //
    // This LUT bellow represent any color component of this following lighting equation :
    //   LUT[(OC << 10) | (C1 << 5) | C2] = ((OC + NC) / 2) << 3
    // with :
    //   NC = (C1 * C2) / 31
    // OC and NC are the colors to be combined.
    // C1 and C2 are the same as in the lighting LUTs.
    extern uint8_t _aun8TLightTable_[32 * 32 * 32];
    //
    // Each LUT bellow represent any color component of this following lighting equation :
    //   LUT[(OC << 10) | (FI << 5) | PC] = ((OC + NC) / 2) << 3
    // with :
    //   NC = ((FI * FC) + ((31 - FI) * PC)) / 31
    // OC and NC are the colors to be combined.
    // FI is the fog's intensity and FC is the fog's color for any component.
    // PC is the panel's color
    extern uint8_t *_pun8TRFogColorTable_;
    extern uint8_t *_pun8TGFogColorTable_;
    extern uint8_t *_pun8TBFogColorTable_;
    //
    //  Each LUT bellow represent any color component of this following lighting equation :
    //    LUT[(OC << 15) | (FI << 10) | (C1 << 5) | C2] = ((OC + NC) / 2) << 3
    //  with :
    //    NC = ((FI * FC) + ((31 - FI) * CL)) / 31
    //    CL = (C1 * C2) / 31
    //  OC and NC are the colors to be combined.
    //  FI is the fog's intensity and FC is the fog's color for any component.
    //  C1 and C2 are the same as in the lighting LUTs.
    extern uint8_t *_pun8TRFogLightTable_;
    extern uint8_t *_pun8TGFogLightTable_;
    extern uint8_t *_pun8TBFogLightTable_;

  } // namespace xprv


  // Safe access to the sine and cosine table
  //!
  //! Returns the sine of the given value,
  //! 0 - 360 degrees are mapped to 0 - (_DEGREECOUNT_ - 1).
  inline float SIN(int32_t n32Val) { return(xprv::_afSinTable_[n32Val & (xprv::_DEGREECOUNT_ - 1)]); }
  //!
  //! Returns the cosine of the given value,
  //! 0 - 360 degrees are mapped to 0 - (_DEGREECOUNT_ - 1).
  inline float COS(int32_t n32Val) { return(xprv::_afCosTable_[n32Val & (xprv::_DEGREECOUNT_ - 1)]); }

  // Faster (but unsafe) access to the sine and cosine table
  //!
  //! Returns the sine of the given value (no range limitter, unsafe),
  //! 0 - 360 degrees are mapped to 0 - (_DEGREECOUNT_ - 1).
  inline float _SIN_(int32_t n32Val) { return(xprv::_afSinTable_[n32Val]); }
  //!
  //! Returns the cosine of the given value (no range limitter, unsafe),
  //! 0 - 360 degrees are mapped to 0 - (_DEGREECOUNT_ - 1).
  inline float _COS_(int32_t n32Val) { return(xprv::_afCosTable_[n32Val]); }

} // namespace vge
#endif
