/** \file MessageDisplayer.h */
// ############################################################################################
//  MessageDisplayer.h
//  Prototype of console-message-displayer functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___MESSAGE_DISPLAYER_H_
#define _BASE___MESSAGE_DISPLAYER_H_

// VortexGE headers
#include "Exception.h"


//!
//! \ingroup Special_Macros
//! \hideinitializer
//! If the macro <b>NDEBUG</b> was defined at the moment MessageDisplayer.h was
//! included, this macro generates no code (does nothing at all). Otherwise, this
//! macro will call VAssertPrint() if the given <b>EXPRESSION</b> evaluates to
//! false (compares equal to zero). Note that MessageDisplayer.h is included from
//! within the VortexGE's main header file (\ref VMain and VMain.h).
//! \n\n
//! <b>Example:</b>
//! \code
//! vassert(radius > 0, "The given radius must be greater than zero");
//! \endcode
//!
//! Note that the <b>MESSAGE</b> can be also NULL like :
//! \code
//! vassert(radius > 0, NULL);
//! \endcode
//! but in this case maybe better to just use the assert() from the standard C
//! library.
//! \n\n
//! Since vassert() is implemented as a macro, if the expression tested has
//! side-effects, program behaviour can be different depending on whether
//! <b>NDEBUG</b> is defined or not. This may create Heisenbugs which go away
//! when debugging is turned on, but (re-)appear on release build.
//!
//! \api stable
#ifdef NDEBUG
#define vassert(EXPRESSION, MESSAGES...) static_cast<void>(0)
#else
#define vassert(EXPRESSION, MESSAGES...) _if_unlikely(!(EXPRESSION)) VAssertPrint(_NFL_, #EXPRESSION, MESSAGES)
#endif


namespace vge {
  //!
  //! \addtogroup MsgDisplayFunctions
  //! @{

  // Support functions
  //!
  //! Call this to clear the console screen.
  //!
  //! \api stable
  extern VGE_EXPORT void VClrScr();
  //!
  //! Redirect the information, warning, error and debugging messages output.
  //!
  //! \api stable
  extern VGE_EXPORT void VRedirectMsgs(FILE *ptInfo = stdout, FILE *ptWarning = stdout, FILE *ptError = stdout, FILE *ptDebug = stdout);
  //!
  //! Enable or disable debugging message.
  //!
  //! \api stable
  extern VGE_EXPORT void VEnableDebugPrint(bool enable = true);
  //!
  //! If the 'abortAfterPrint' is set to 'true', the next call to VFatalError()
  //! will print the error message and then abort the application (instead of
  //! throwing the CFatalError exception). Use this function with caution !!!
  //!
  //! \api stable
  extern VGE_EXPORT void VAbortOnFatalError(bool abortAfterPrint = false);

  // Fatal error message displayer
  //!
  //! Display fatal error message and throw the appropriate exception.
  //! This function will print to the I/O stream specified in the 'ptError'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VFatalError(EERRCode eCode, const char *pcMsg, ...) _attr_format_printf(2) _attr_noreturn;
  //!
  //! Display fatal error message and throw the appropriate exception.
  //! This function will print to the I/O stream specified in the 'ptError'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VFatalError(EERRCode eCode, const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...) _attr_format_printf(5) _attr_noreturn;

  // Error message displayer
  //!
  //! This function will print to the I/O stream specified in the 'ptError'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //! Display error message and throw the appropriate exception.
  //!
  //! \api stable
  extern VGE_EXPORT void VError(EERRCode eCode, const char *pcMsg, ...)  _attr_format_printf(2) _attr_noreturn;
  //!
  //! Display error message and throw the appropriate exception.
  //! This function will print to the I/O stream specified in the 'ptError'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VError(EERRCode eCode, const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...) _attr_format_printf(5) _attr_noreturn;

  // Warning message displayer
  //!
  //! Display warning message.
  //! This function will print to the I/O stream specified in the 'ptWarning'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VWarning(const char *pcMsg, ...) _attr_format_printf(1);
  //!
  //! Display warning message.
  //! This function will print to the I/O stream specified in the 'ptWarning'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VWarning(const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...) _attr_format_printf(4);

  // Information message displayer
  //!
  //! Display information message.
  //! This function will print to the I/O stream specified in the 'ptInfo'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VInfo(const char *pcMsg, ...) _attr_format_printf(1);
  //!
  //! Display information message.
  //! This function will print to the I/O stream specified in the 'ptInfo'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VInfo(const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...) _attr_format_printf(4);
  //!
  //! Display custom information message.
  //! This function will print to the I/O stream specified in the 'ptInfo'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VPrint(const char *pcMsg, ...) _attr_format_printf(1);

  // Debug message displayer
  //!
  //! Display debugging message, affected by VEnableDebugPrint().
  //! This function will print to the I/O stream specified in the 'ptDebug'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //!
  //! \api stable
  extern VGE_EXPORT void VDebugPrint(const char *pcMsg, ...) _attr_format_printf(1);

  // Assertion message displayer
  //!
  //! Display the assertion text (the test code) and the assertion message and
  //! then abort the application. If the assertion message ('pcMsg') is NULL,
  //! only the assertion text which will be displayed.
  //! \n\n
  //! This function will print to the I/O stream specified in the 'ptDebug'
  //! parameter in the VRedirectMsgs() function call (the default is stdout).
  //! This function is not affected by VEnableDebugPrint().
  //! \n\n
  //! It is not reccomended to call this function directly; use the vassert()
  //! macro instead.
  //!
  //! \api stable
  extern VGE_EXPORT void VAssertPrint(const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcAssertionText, const char *pcMsg, ...) _attr_format_printf(5) _attr_noreturn;

  //! @}
  //!
} // namespace vge
#endif
