/** \file SpecialClasses.h */
// ############################################################################################
//  SpecialClasses.h
//  Definition & implementation of some special purpose classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___SPECIAL_CLASSES_H_
#define _BASE___SPECIAL_CLASSES_H_

// VortexGE headers
#include "VBasics.h"

//!
//! \addtogroup Special_Macros
//! @{

//!
//! \hideinitializer
//! Use this macro to make a class become underiveable.
//!
//! \api stable
#define VCLASS_UNDERIVABLE virtual private CUnDerivable

//!
//! \hideinitializer
//! Use this macro to make a class become uncopyable.
//!
//! \api stable
#define VCLASS_UNCOPYABLE virtual public CUnCopyable

//!
//! \hideinitializer
//! Use this macro to make a class become "can abort on destruction".
//!
//! \api stable
#define VCLASS_CAN_ABORT_ON_DESTRUCTION virtual public CCanAbortOnDestruction
//!
//! \hideinitializer
//! Use this macro to make a class become "can abort on destruction".
//!
//! \api stable
#define VCLASS_CAN_ABORT_ON_DESTRUCTION_CTOR_BEGIN(CTYPE) CCanAbortOnDestruction::_ctype = CTYPE;

//!
//! \hideinitializer
//! Use this macro to make a class become singleton.
//!
//! \api pretty stable
#define VCLASS_SINGLETON VCLASS_UNDERIVABLE
//!
//! \hideinitializer
//! Use this macro to make a class become singleton.
//!
//! \api pretty stable
#define VCLASS_SINGLETON_DECLARE(CLASS) \
  public :                              \
    static CLASS& Instance();
//!
//! \hideinitializer
//! Use this macro to make a class become singleton.
//!
//! \api pretty stable
#define VCLASS_SINGLETON_IMPLEMENT(CLASS) \
  CLASS& CLASS::Instance()                \
  {                                       \
    static CLASS csInst;                  \
    return(csInst);                       \
  }

//!
//! \hideinitializer
//! Use this macro to implement a class wrapper for enumeration values.
//! Using this class, two or more enumerations can have the same symbol names.
//!
//! \api beta
#define VCLASS_ENUM_WRAPPER_IMPLEMENT(CLASS, ENUM_VALUES...)  \
  class CLASS {                                               \
    public:                                                   \
      enum _ENUM { ENUM_VALUES };                             \
      inline          CLASS()                              {} \
      inline          CLASS(const CLASS &v) : _v(v._v    ) {} \
      inline          CLASS(      _ENUM  v) : _v(v       ) {} \
      inline explicit CLASS(        int  v) : _v(_ENUM(v)) {} \
      inline operator _ENUM() const { return(_v); }           \
    private:                                                  \
      _ENUM _v;                                               \
  };

//!
//! \hideinitializer
//! Use this macro to implement a lock class.\n\n
//! <b>Example:</b>
//! \code
//! class TestClass {
//!   public:
//!     TestClass() { /*  */ }
//!
//!     void SomeFunction() {
//!       Lock1 locker(this);
//!
//!       /* Do something  */
//!     }
//!
//!     void SomeOtherFunction() {
//!       Lock2 locker(this);
//!
//!       /* Do something else  */
//!     }
//!
//!     void lock()   { _mutex1.Lock();   }
//!     void unlock() { _mutex1.Unlock(); }
//!
//!   private:
//!            CMutex _mutex1;
//!     static CMutex _mutex2;
//!
//!     VCLASS_LOCK_IMPLEMENT(Lock1, TestClass, lock(), unlock())
//!     VCLASS_LOCK_IMPLEMENT(Lock2, TestClass, _mutex2.Lock(), _mutex2.Unlock())
//! };
//! \endcode
//!
//! \api beta
#define VCLASS_LOCK_IMPLEMENT(CLASS, TARGET, LOCKER, UNLOCKER) \
  class CLASS {                                                \
    private:                                                   \
      TARGET *_inst;                                           \
    public:                                                    \
      CLASS(TARGET *inst) : _inst( inst) { _inst->LOCKER;   }  \
      CLASS(TARGET &inst) : _inst(&inst) { _inst->LOCKER;   }  \
      ~CLASS() throw()                   { _inst->UNLOCKER; }  \
  };

//! @}
//!

namespace vge {

  //!
  //! \brief The parent of all underivable classes
  //!
  //! \note Instead of using this class directly, use the macro \ref VCLASS_UNDERIVABLE.
  //!
  //! \api stable
  class VGE_EXPORT CUnDerivable {
    protected:
      inline CUnDerivable() {}
      inline ~CUnDerivable() throw() {}
  };


  //!
  //! \brief The parent of all uncopyable classes
  //!
  //! \note Instead of using this class directly, use the macro \ref VCLASS_UNCOPYABLE.
  //!
  //! \api stable
  class VGE_EXPORT CUnCopyable {
    public:
      inline CUnCopyable() {}
      inline ~CUnCopyable() throw() {}

    private:
      CUnCopyable (const CUnCopyable&);
      CUnCopyable &operator=(const CUnCopyable&);
  };


  //!
  //! \brief The parent of all "can abort on destruction" classes
  //!
  //! One should not use this class directly unless wants to
  //! implement one's own media manager/pool class.
  //!
  //! \note Instead of using this class directly, use the macro \ref VCLASS_CAN_ABORT_ON_DESTRUCTION.
  //!
  //! \api beta
  class VGE_EXPORT CCanAbortOnDestruction {
    public:
      //!
      //! Construct a new instance using the given setting.
      CCanAbortOnDestruction(bool abortOnDestruction = false);
      //!
      //! Standard dtor.
      virtual ~CCanAbortOnDestruction() throw();

      // Policy functions
      //!
      //! Make the class abort the application if the class instance
      //! is destructed.
      inline void MakeItAbortOnDestruction(bool abortOnDestruction)
      { _abod = abortOnDestruction; }
      //!
      //! Check if the class instance will make the application abort
      //! on destruction.
      inline bool WillAbortOnDestruction() const
      { return(_abod); }
      //!
      //! Returns the type name of the class
      const string &GetClassTypeName() const
      { return(_ctype); }

    protected:
      bool   _abod;  // If 'true', the applciation will abort if this class instance is destroyed
      string _ctype; // Type name of the class
  };

} // namespace vge
#endif
