/** \file ApplicationObject.h */
// ############################################################################################
//  ApplicationObject.h
//  Definition & implementation of CBasicAppObjects class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _HIGH_LEVEL_CLASS___APPLICATION_OBJECT_H_
#define _HIGH_LEVEL_CLASS___APPLICATION_OBJECT_H_

// VortexGE headers
#include "../Base/MessageCatalog.h"
#include "EventHandler.h"


namespace vge {

  // Foward reference from "lib/3DEnvironment/3DEnvironment.h"
  class C3DEnv;

  // Foward reference from "lib/AudioEnvironment/AudioEnvironment.h"
  class CAudioEnv;

  // Foward reference from "lib/AudioEnvironment/MidiPlayer.h"
  class CMidiPlayer;


  //!
  //! \brief Standard display mode
  //!
  //! \api pretty stable
  enum ESDM {
    ESDM_320X240_NORMAL, //!< Window size 320x240 pixels, render buffer size 320x240 pixels.
    ESDM_320X240_LOWRES, //!< Window size 320x240 pixels, render buffer size 160x120 pixels.
    ESDM_640X480_NORMAL, //!< Window size 640x480 pixels, render buffer size 640x480 pixels.
    ESDM_640X480_LOWRES, //!< Window size 640x480 pixels, render buffer size 320x240 pixels.
    ESDM_800X600_NORMAL, //!< Window size 800x600 pixels, render buffer size 800x600 pixels.
    ESDM_800X600_LOWRES  //!< Window size 800x600 pixels, render buffer size 400x300 pixels.
  };

  //!
  //! \brief Audio initialization mode
  //!
  //! \api pretty stable
  enum EAIM {
    EAIM_Normal,        //!< Try to initialize, if failed, ask the user for the next action.
    EAIM_NormalNoRetry, //!< Try to initialize, if failed, disable audio.
    EAIM_Enabled,       //!< Try to initialize, if failed, throw exception.
    EAIM_Disabled       //!< Don't initialize audio.
  };

  //!
  //! \brief 3D initialization mode
  //!
  //! \api pretty stable
  enum E3DIM {
    E3DIM_Normal,   //!< Initialize 3D automatically.
    E3DIM_Manual,   //!< Manually initialize 3D.
    E3DIM_Disabled, //!< Disable 3D completely. Using CBasicAppObjects::i3DEnv() and using rendering functions from CBasicAppObjects::iVideoEnv() will cause undefined behavior.
  };


  //!
  //! \brief A structure used for storing mouse settings
  //!
  //! \api beta
  struct VGE_EXPORT TMouseSettings {
    EBtnSymbol leftClickButton;   //!< Button symbol to be used to for "left click" button.
    EBtnSymbol middleClickButton; //!< Button symbol to be used to for "middle click" button.
    EBtnSymbol rightClickButton;  //!< Button symbol to be used to for "right click" button.
    EBtnSymbol scrollUpButton;    //!< Button symbol to be used to for "scroll up" button.
    EBtnSymbol scrollDownButton;  //!< Button symbol to be used to for "scroll down" button.
    uint32_t   doubleClickDelay;  //!< Double click delay (in milliseconds).

    //!
    //! Construct a default mouse settings for right-handed people.
    inline TMouseSettings(EBtnSymbol leftClickButton_   = VBS_MB1,
                          EBtnSymbol middleClickButton_ = VBS_MB2,
                          EBtnSymbol rightClickButton_  = VBS_MB3,
                          EBtnSymbol scrollUpButton_    = VBS_MB4,
                          EBtnSymbol scrollDownButton_  = VBS_MB5,
                          uint32_t   doubleClickDelay_  = 400)
    : leftClickButton  (leftClickButton_),
      middleClickButton(middleClickButton_),
      rightClickButton (rightClickButton_),
      scrollUpButton   (scrollUpButton_),
      scrollDownButton (scrollDownButton_),
      doubleClickDelay (doubleClickDelay_)
    {}

    // Initialization functions
    //!
    //! Change the global mouse settings to the default settings for
    //! left-handed people.
    inline static void UseTheDefaultLeftHandedMouseSettings()
    { GLOBAL_SET = TMouseSettings(VBS_MB3, VBS_MB2, VBS_MB1); }
    //!
    //! Change the global mouse settings to the default settings for
    //! right-handed people.
    inline static void UseTheDefaultRightHandedMouseSettings()
    { GLOBAL_SET = TMouseSettings(); }
    //!
    //! Change the global mouse settings with the given settings.
    inline static void ChangeGlobalMouseSettings(const TMouseSettings &newSettings)
    { GLOBAL_SET = newSettings; }

    // Data
    static TMouseSettings GLOBAL_SET; //!< The global mouse settings (by default it is for right-handed people).
  };


  //!
  //! \brief A structure used for determining the global key settings
  //!
  //! \api beta
  struct VGE_EXPORT TKeySettings {
    uint32_t keyRepeatStart; //!< Key repeat start delay (in milliseconds).
    uint32_t keyRepeatDelay; //!< Key repeat delay (in milliseconds).

    //!
    //! Construct a default key settings.
    inline TKeySettings(uint32_t keyRepeatStart_ = 240,
                        uint32_t keyRepeatDelay_ = 20)
    : keyRepeatStart(keyRepeatStart_),
      keyRepeatDelay(keyRepeatDelay_)
    {}

    // Initialization functions
    //!
    //! Change the global key settings to the default settings.
    inline static void UseTheDefaultKeySettings()
    { GLOBAL_SET = TKeySettings(); }
    //!
    //! Change the global key settings with the given settings.
    inline static void ChangeGlobalKeySettings(const TKeySettings &newSettings)
    { GLOBAL_SET = newSettings; }

    // Data
    static TKeySettings GLOBAL_SET; //!< The global key settings
  };


  //!
  //! \brief A class containing all basic objects needed by most VortexGE applications
  //! 
  //! Affecting environment variables :
  //!   - VGE_MESSAGE_CATALOG_PATH
  //!   - VGE_LOCALE
  //!   - LC_MESSAGES
  //!   - LC_ALL
  //!   - LANGUAGE
  //!   - LANG
  //!
  //! See the section "Environment Variables" form the main doumentation (not the
  //! API documentation) for more details.
  //!
  //! \api beta
  class VGE_EXPORT CBasicAppObjects {
    public:
      //!
      //! Initialize the application using the given parameters.
      //! \param[in] utf8WindowTitle    Text to be displayed on the window's title bar.
      //! \param[in] windowIconFileName XPM file to be used as the window's icon.
      //! \param[in] windowWidth        Width of the window (zero means use the default width of 640 pixels).
      //! \param[in] windowHeight       Height of the window (zero means use the default height of 480 pixels).
      //! \param[in] bufferWidth        Width of the render buffer (zero means the same as the window's width).
      //! \param[in] bufferHeight       Height of the render buffer (zero means the same as the window's height).
      //! \param[in] eAudioInitMode      See the vge::EAIM enumeration.
      //! \parampin] e3DInitMode         See the vge::E3DIM enumeration.
      CBasicAppObjects(const char *utf8WindowTitle = NULL, const char *windowIconFileName = NULL,
                       uint32_t    windowWidth     =    0, uint32_t    windowHeight       =    0,
                       uint32_t    bufferWidth     =    0, uint32_t    bufferHeight       =    0,
                       EAIM        eAudioInitMode  = EAIM_Normal,
                       E3DIM       e3DInitMode     = E3DIM_Normal);
      //!
      //! Initialize the application using the given parameters.
      //! \param[in] eDisplayMode       See the vge::ESDM enumeration.
      //! \param[in] utf8WindowTitle    Text to be displayed on the window's title bar.
      //! \param[in] windowIconFileName XPM file to be used as the window's icon.
      //! \param[in] eAudioInitMode     See the vge::EAIM enumeration.
      //! \parampin] e3DInitMode        See the vge::E3DIM enumeration.
      CBasicAppObjects(ESDM        eDisplayMode       = ESDM_640X480_LOWRES,
                       const char *utf8WindowTitle    = NULL,
                       const char *windowIconFileName = NULL,
                       EAIM        eAudioInitMode     = EAIM_Normal,
                       E3DIM       e3DInitMode        = E3DIM_Normal);
      ~CBasicAppObjects() throw();

      // Internationalization functions
      //!
      //! Returns the current system locale. Affected by the environment variables
      //! VGE_LOCALE / LC_MESSAGES / LC_ALL / LANGUAGE / LANG.
      static const char *SystemLocale();
      //!
      //! Initialize the internationalization support (needn't be called manually
      //! as long as the user always use VortexGE's high level API).
      //! \param[in] pathToMessageCatalog Path to the directory containing the catalog 
      //!                                 files (NULL means using the default directory
      //!                                 for the library's internal message catalog).
      //! \param[in] messagesLocale       Locale name for determining which catalog file
      //!                                 to be loaded from the catalog directory (NULL
      //!                                 means using the current system locale returned
      //!                                 by SystemLocale().
      static void InitializeI18N(const char *pathToMessageCatalog = NULL, const char *messagesLocale = NULL);
      //!
      //! Set the character to be used as the decimal separator.
      //! \n\n
      //! Generally it shouldn't be called manually since InitializeI18N()
      //! should be able to set the correct separator.
      inline static void SetDecimalSeparator(char sep)
      { _decSep = sep; }
      //!
      //! Returns the character used as the decimal separator.
      inline static char DecimalSeparator() 
      { return(_decSep); }
      //!
      //! Returns the application's CMessageCatalog instance.\n\n
      //! One can load (and append) his own message catalog files by using code like:
      //! \code
      //! CBasicAppObjects::MessageCatalog().LoadCatalog("path/to/user/dir", NULL, true);
      //! \endcode
      //! Then one can translate message using code like:
      //! \code
      //! const char *translatedMsg = CBasicAppObjects::MessageCatalog().LookUp("my message");
      //! \endcode
      //! or:
      //! \code
      //! const char *translatedMsg = _M("my message");
      //! \endcode
      inline static CMessageCatalog &MessageCatalog()
      { return(CBasicAppObjects::_msgCat); }

      // Environment accessor
      //!
      //! Returns the application's CAudioEnv instance.
      CAudioEnv &iAudioEnv()
      { return(_csAudioEnv); }
      //!
      //! Returns the application's CMidiPlayer instance.
      CMidiPlayer &iMidiPlayer()
      { return(_csMidiPlayer); }
      //!
      //! Returns the application's CVideoEnv instance.
      CVideoEnv &iVideoEnv() 
      { return(*_csVideoEnv); }
      //!
      //! Returns the application's C3DEnv instance.
      C3DEnv &i3DEnv()
      { return(*_cs3DEnv); }

      // Information functions
      //!
      //! Check if the audio has been disabled in the initialization (thus
      //! making the instance returned by iAudioEnv() and iMidiPlayer() unuseable).
      bool AudioIsDisabled() const
      { return(_bAudioDisabled); }
      //!
      //! Check if the midi player has been disabled in the initialization
      //! (thus making the instance returned by iMidiPlayer() unuseable).
      bool MidiIsDisabled() const
      { return(_bMidiDisabled); }

    protected:
      // Initialization functions
      void _Initialize_(const char *utf8WindowTitle, const char *pucWindowIconFileName ,
                        uint32_t    un32WindowWidth, uint32_t    un32WindowHeight,
                        uint32_t    un32BufferWidth, uint32_t    un32BufferHeight,
                        EAIM        eAudioInitMode,
                        E3DIM       e3DInitMode);

      // Data
      static char            _decSep; // Decimal separator
      static CMessageCatalog _msgCat; // Message catalog to be used by the library
                                      // when displaying dialog messages
      CAudioEnv   &_csAudioEnv;
      CMidiPlayer &_csMidiPlayer;
      CVideoEnv   *_csVideoEnv;
      C3DEnv      *_cs3DEnv;
      bool         _bAudioDisabled;
      bool         _bMidiDisabled;
  };


  // Convenience functions
  //!
  //! Convenience function that can be used when translating messages using
  //! the static CMessageCatalog instance of the CBasicAppObjects class.
  //! \n\n
  //! If no translation is found, the original pointer (const char *message)
  //! will be returned back.
  inline const char *_M(const char *message)
  { return(CBasicAppObjects::MessageCatalog().LookUp(message)); }
  //!
  //! Safe convenience function that can be used when translating messages using
  //! the static CMessageCatalog instance of the CBasicAppObjects class.
  //! \n\n
  //! Returns a copy of the translated message (or a copy of the original message
  //! if no translation is found).
  inline const string _S(const string &message)
  { return(CBasicAppObjects::MessageCatalog().LookUp(message.c_str())); }

} // namespace vge
#endif
