// ############################################################################################
//  EventHandler.cc
//  Implementation of CEventHandler and CEventHandlerExt classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../VideoEnvironment/VideoEnvironment.h"
#include "EventHandler.h"


namespace vge {

//
// The base of all event handler classess
//
CEventHandlerBase::CEventHandlerBase(CVideoEnv *pcsVidEnv)
: _pcsVidEnv(pcsVidEnv), _bDoExit(false)
{
  _bActive  = pcsVidEnv->IsActive();
  _bVisible = pcsVidEnv->IsVisible();
}

CEventHandlerBase::~CEventHandlerBase() throw()
{
}

bool CEventHandlerBase::EventMainLoopEnter()
{
  return(true);
}

bool CEventHandlerBase::EventMainLoopLeave()
{
  return(true);
}

void CEventHandlerBase::EventKeyPress(EKeySymbol eKeySym)
{
  UNUSED_PARAM(eKeySym);
}

void CEventHandlerBase::EventKeyPress(EKeySymbol eKeySym, const char *str)
{
  UNUSED_PARAMS_2(eKeySym, str);
}

void CEventHandlerBase::EventKeyRelease(EKeySymbol eKeySym)
{
  UNUSED_PARAM(eKeySym);
}

void CEventHandlerBase::EventKeyRelease(EKeySymbol eKeySym, const char *str)
{
  UNUSED_PARAMS_2(eKeySym, str);
}

void CEventHandlerBase::EventButtonPress(EBtnSymbol eButtonSym)
{
  UNUSED_PARAM(eButtonSym);
}

void CEventHandlerBase::EventButtonPress(EBtnSymbol eButtonSym, int32_t n32XPos, int32_t n32YPos)
{
  UNUSED_PARAMS_3(eButtonSym, n32XPos, n32YPos);
}

void CEventHandlerBase::EventButtonRelease(EBtnSymbol eButtonSym)
{
  UNUSED_PARAM(eButtonSym);
}

void CEventHandlerBase::EventButtonRelease(EBtnSymbol eButtonSym, int32_t n32XPos, int32_t n32YPos)
{
  UNUSED_PARAMS_3(eButtonSym, n32XPos, n32YPos);
}

void CEventHandlerBase::EventMouseMove(int32_t n32XPos, int32_t n32YPos)
{
  UNUSED_PARAMS_2(n32XPos, n32YPos);
}

void CEventHandlerBase::EventMouseMove(int32_t n32XPos, int32_t n32YPos, bool bButtonState[5])
{
  UNUSED_PARAMS_3(n32XPos, n32YPos, bButtonState);
}

void CEventHandlerBase::EventUpdateFrame(bool bWindowIsActive)
{
  UNUSED_PARAM(bWindowIsActive);
}

void CEventHandlerBase::EventWindowVisibility(bool bWindowIsVisible)
{
  UNUSED_PARAM(bWindowIsVisible);
}

void CEventHandlerBase::EventWindowActivated()
{
  _pcsVidEnv->AutoRepeatOff();
}

void CEventHandlerBase::EventWindowDeactivated()
{
  _pcsVidEnv->AutoRepeatOn();
}

void CEventHandlerBase::EventCloseWindow()
{
  _bDoExit = true;
}

void CEventHandlerBase::EventPasteData(CClipboard::CDT eDataType)
{
  UNUSED_PARAM(eDataType);
}


//
// The standard event handler class
//
CEventHandler::CEventHandler(CVideoEnv *pcsVidEnv)
: CEventHandlerBase(pcsVidEnv)
{
}

// NOTE: * The default DoMainLoop() will return -1 if error.
//       * When you override this function, make sure that you call :
//           1. EventMainLoopEnter()
//           2. EventMainLoopLeave()
//         in the correct order (unless you really want it differently).
int32_t CEventHandler::DoMainLoop()
{
  // Invoke the "EventMainLoopEnter" event
  if(!EventMainLoopEnter()) return(-1);

  // Main loop
  while(!Exiting()) {
    if(PumpEvents()) HandleEvents();
  }

  // Invoke the "EventMainLoopLeave" event
  if(!EventMainLoopLeave()) return(-1);

  // OK !
  return(0);
}

bool CEventHandler::PumpEvents()
{
  _if_likely(_pcsVidEnv->IsActive())
    _pcsVidEnv->DispatchEvent();
  else
    _pcsVidEnv->BlockedDispatchEvent();
  return(true);
}

void CEventHandler::HandleEvents()
{
  // The window is active
  if(_pcsVidEnv->IsActive()) {
    // Check changes in window states
    if(!_bActive) {
      EventWindowActivated();
      _bActive = true;
    }
    if(_pcsVidEnv->IsVisible() != _bVisible) {
      _bVisible = _pcsVidEnv->IsVisible();
      EventWindowVisibility(_bVisible);
    }
    // Check other events
    if(_pcsVidEnv->CheckEvent()) {
      // Keyboard events
      if (_pcsVidEnv->EventKeyPress()) {
        char           kcb[32];
        const uint32_t n = _pcsVidEnv->GetKeyString(kcb, sizeof(kcb));
        kcb[n] = 0;
        EventKeyPress(_pcsVidEnv->GetKeySym());
        EventKeyPress(_pcsVidEnv->GetKeySym(), kcb);
      }
      else if(_pcsVidEnv->EventKeyRelease()) {
        char           kcb[32];
        const uint32_t n = _pcsVidEnv->GetKeyString(kcb, sizeof(kcb));
        kcb[n] = 0;
        EventKeyRelease(_pcsVidEnv->GetKeySym());
        EventKeyRelease(_pcsVidEnv->GetKeySym(), kcb);
      }
      // Mouse events
      else if(_pcsVidEnv->EventButtonPress()) {
        EventButtonPress(_pcsVidEnv->GetButtonSym());
        EventButtonPress(_pcsVidEnv->GetButtonSym(), _pcsVidEnv->GetButtonX(), _pcsVidEnv->GetButtonY());
      }
      else if(_pcsVidEnv->EventButtonRelease()) {
        EventButtonRelease(_pcsVidEnv->GetButtonSym());
        EventButtonRelease(_pcsVidEnv->GetButtonSym(), _pcsVidEnv->GetButtonX(), _pcsVidEnv->GetButtonY());
      }
      else if(_pcsVidEnv->EventMouseMove()) {
        bool bBS[5] = { _pcsVidEnv->GetButton1State(),
                        _pcsVidEnv->GetButton2State(),
                        _pcsVidEnv->GetButton3State(),
                        _pcsVidEnv->GetButton4State(),
                         _pcsVidEnv->GetButton5State() };
        EventMouseMove(_pcsVidEnv->GetMouseX(), _pcsVidEnv->GetMouseY());
        EventMouseMove(_pcsVidEnv->GetMouseX(), _pcsVidEnv->GetMouseY(), bBS);
      }
      // Close window event
      else if(_pcsVidEnv->EventCloseWindow()) {
        EventCloseWindow();
      }
      // Paste data event
      else if(_pcsVidEnv->EventPasteData()) {
        EventPasteData(_pcsVidEnv->GetPastedDataType());
      }
      // Unknown events
      else {
      }
    }
    // Update frame only if the window is visible
    if(_pcsVidEnv->IsVisible()) EventUpdateFrame(true);
  }
  // The window is inactive
  else {
    // Check changes in window states
    if(_bActive) {
      EventWindowDeactivated();
      EventUpdateFrame(false);
      _bActive = false;
    }
    if(_pcsVidEnv->IsVisible() != _bVisible) {
      _bVisible = _pcsVidEnv->IsVisible();
      EventWindowVisibility(_bVisible);
      EventUpdateFrame(false);
    }
    // Check other events
    if(_pcsVidEnv->CheckEvent()) {
      // Check for close window event
      if(_pcsVidEnv->EventCloseWindow())
        EventCloseWindow();
      // Check for expose event and update the frame
      else if(_pcsVidEnv->EventExpose())
        EventUpdateFrame(false);
    }
  }
}


//
// The extended event handler class
//
CEventHandlerExt::CEventHandlerExt(CVideoEnv *pcsVidEnv, float fWantedFPS)
: CEventHandlerBase(pcsVidEnv), _bUpdateFPSInfoString(true), _bDisableFPSControl(false)
{
  _csFPSCtrl.SetWantedFPS(fWantedFPS);
}

// NOTE: * The default DoMainLoop() will return the average FPS or -1 if error.
//       * When you override this function, make sure that you call :
//           1. EventMainLoopEnter()
//           2. SaveStartTicks()
//           3. SaveEndTicks();
//           4. EventMainLoopLeave()
//         in the correct order (unless you really want it differently).
int32_t CEventHandlerExt::DoMainLoop()
{
  // Invoke the "EventMainLoopEnter" event
  if(!EventMainLoopEnter()) return(-1);

  // Main loop
  SaveStartTicks();
  while(!Exiting()) {
    if(PumpEvents()) HandleEvents();
  }
  SaveEndTicks();

  // Invoke the "EventMainLoopLeave" event
  if(!EventMainLoopLeave()) return(-1);

  // Return the average frame rate (x 100)
  return(int32_t(GetAvgFPS() * 100.0f));
}

bool CEventHandlerExt::PumpEvents()
{
  if(_csFPSCtrl.PumpEventAndDelay(*_pcsVidEnv, _bDisableFPSControl)) return(false);
  if(_csFPSCtrl.AdjustDelay(_bDisableFPSControl)) {
    if(_bUpdateFPSInfoString) {
      snprintf(_acFPSBuff, sizeof(_acFPSBuff), "%05.1f fps, %06d delays",
               _csFPSCtrl.GetFPS(),
               _csFPSCtrl.GetDelayFactor());
    }
  }
  return(true);
}

void CEventHandlerExt::HandleEvents()
{
  // The window is active
  if(_pcsVidEnv->IsActive()) {
    // Check changes in window states
    if(!_bActive) {
      EventWindowActivated();
      _bActive = true;
    }
    if(_pcsVidEnv->IsVisible() != _bVisible) {
      _bVisible = _pcsVidEnv->IsVisible();
      EventWindowVisibility(_bVisible);
    }
    // Check other events
    if(_pcsVidEnv->CheckEvent()) {
      // Keyboard events
      if (_pcsVidEnv->EventKeyPress()) {
        EventKeyPress(_pcsVidEnv->GetKeySym());
      }
      else if(_pcsVidEnv->EventKeyRelease()) {
        EventKeyRelease(_pcsVidEnv->GetKeySym());
      }
      // Mouse events
      else if(_pcsVidEnv->EventButtonPress()) {
        EventButtonPress(_pcsVidEnv->GetButtonSym());
        EventButtonPress(_pcsVidEnv->GetButtonSym(), _pcsVidEnv->GetButtonX(), _pcsVidEnv->GetButtonY());
      }
      else if(_pcsVidEnv->EventButtonRelease()) {
        EventButtonRelease(_pcsVidEnv->GetButtonSym());
        EventButtonRelease(_pcsVidEnv->GetButtonSym(), _pcsVidEnv->GetButtonX(), _pcsVidEnv->GetButtonY());
      }
      else if(_pcsVidEnv->EventMouseMove()) {
        bool bBS[5] = { _pcsVidEnv->GetButton1State(),
                        _pcsVidEnv->GetButton2State(),
                        _pcsVidEnv->GetButton3State(),
                        _pcsVidEnv->GetButton4State(),
                         _pcsVidEnv->GetButton5State() };
        EventMouseMove(_pcsVidEnv->GetMouseX(), _pcsVidEnv->GetMouseY());
        EventMouseMove(_pcsVidEnv->GetMouseX(), _pcsVidEnv->GetMouseY(), bBS);
      }
      // Close window event
      else if(_pcsVidEnv->EventCloseWindow()) {
        EventCloseWindow();
      }
      // Paste data event
      else if(_pcsVidEnv->EventPasteData()) {
        EventPasteData(_pcsVidEnv->GetPastedDataType());
      }
      // Unknown events
      else {
      }
    }
    // Update frame only if the delay counter reaches zero and the window is visible
    if((_bDisableFPSControl || (_csFPSCtrl.GetDelayCounter() == 0)) &&
       _pcsVidEnv->IsVisible()) EventUpdateFrame(true);
  }
  // The window is inactive
  else {
    // Check changes in window states
    if(_bActive) {
      EventWindowDeactivated();
      EventUpdateFrame(false);
      _bActive = false;
    }
    if(_pcsVidEnv->IsVisible() != _bVisible) {
      _bVisible = _pcsVidEnv->IsVisible();
      EventWindowVisibility(_bVisible);
      EventUpdateFrame(false);
    }
    // Check other events
    if(_pcsVidEnv->CheckEvent()) {
      // Check for close window event
      if(_pcsVidEnv->EventCloseWindow())
        EventCloseWindow();
      // Check for expose event and update the frame
      else if(_pcsVidEnv->EventExpose())
        EventUpdateFrame(false);
    }
  }
}

} // namespace vge
