/** \file Archiver.h */
// ############################################################################################
//  Archiver.h
//  Definition & implementation of CArchiver and CDeArchiver classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _SYSTEM___ARCHIVER_H_
#define _SYSTEM___ARCHIVER_H_

// VortexGE headers
#include "../System/DataSinkSource.h"


namespace vge {

  //!
  //! \brief A class used for archiving files (without compression support)
  //!
  //! The size of the input files and the archive file must not exceed the
  //! capacity of uint32_t (4GB).
  //!
  //! \api experiment
  class VGE_EXPORT CArchiver : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct a new CArchiver instance using the given parameters.
      CArchiver(const string &inputFilesBasePath = "");
      //!
      //! Construct a new CArchiver instance using the given parameters.
      CArchiver(const string &inputFilesBasePath, const vector<string> &inputFiles);
      //!
      //! Construct a new CArchiver instance using the given parameters.
      CArchiver(const string &inputFilesBasePath, const vector<string> &inputFiles, const vector<char> &userData);

      // Initialization functions
      //!
      //! Clear all previously set initialization data.
      void Clear();
      //!
      //! Set the base path (prefix) for all input files to be archived.
      inline void SetInputFilesBasePath(const string &basePath)
      {
        _inputFilesBasePath = basePath;
        _FixInputFilesBasePath();
      }
      //!
      //! Set the input files to be archived.
      inline void SetInputFiles(const vector<string> &inputFiles)
      { _inputFiles = inputFiles; }
      //!
      //! Add input files to be archived.
      inline void AddInputFiles(const vector<string> &inputFiles)
      { _inputFiles.insert(_inputFiles.end(), inputFiles.begin(), inputFiles.end()); }
      //!
      //! Add an input file to be archived.
      inline void AddInputFile(const string &inputFile)
      { _inputFiles.push_back(inputFile); }
      //!
      //! Set additional user data to be included in the archive.
      inline void SetUserData(const vector<char> &userData)
      { _userData = userData; }
      //!
      //! Set additional user data to be included in the archive.
      void SetUserDataStr(const string &userData);

      // Archiving functions
      //!
      //! Archive the source files.
      //! \param[in] sink The name of the destination archive file (*.vaf) or
      //!                 an instance of ostream derived class.
      void ProcessFiles(const CDataSink &sink) const;

    protected:
      // Helper functions
      inline void _FixInputFilesBasePath()
      {
        if(_inputFilesBasePath.empty()) return;
        if(_inputFilesBasePath[_inputFilesBasePath.length() - 1] != '/') _inputFilesBasePath += '/';
      }

      // Data
      string         _inputFilesBasePath; // Base path for all source (input) files
      vector<string> _inputFiles;         // List of input files
      vector<char>   _userData;           // User data
  };

  //!
  //! \brief A class used for de-archiving archives (without compression support)
  //!
  //! The size of the archive file and output files must not exceed the
  //! capacity of uint32_t (4GB).
  //!
  //! \api experiment
  class VGE_EXPORT CDeArchiver : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct a new CDeArchiver instance using the given parameters.
      CDeArchiver(const string &outputFilesBasePath = "");
      //!
      //! Standard dtor.
      virtual ~CDeArchiver() throw();

      // Initialization functions
      //!
      //! Clear all previously set initialization and extracted data.
      void Clear();
      //!
      //! Set the base path (prefix) for all output files to be extracted.
      inline void SetTargetFilesBasePath(const string &basePath)
      {
        _outputFilesBasePath = basePath;
        _FixOutputFilesBasePath();
      }

      // De-archiving functions
      //!
      //! De-archive the source files.
      //! \param[in] source The name of the source archive file (*.vaf) or
      //!            an instance of istream derived class.
      void ProcessFiles(const CDataSource &source);
      //!
      //! Get the name of the output (extracted) files.
      inline const vector<string> &GetOutputFiles() const
      { return(_outputFiles); }
      //!
      //! Get the user data.
      inline const vector<char> &GetUserData() const
      { return(_userData); }
      //!
      //! Get the user data.
      const string GetUserDataStr() const;
      //!
      //! This virtual function will be called after the CDeArchiver got
      //! the user data chunk from the archive file. Override this function
      //! if user data validation is needed.
      virtual void ValidateUserData(const vector<char>& userData) const;

    protected:
      // Helper functions
      inline void _FixOutputFilesBasePath()
      {
        if(_outputFilesBasePath.empty()) return;
        if(_outputFilesBasePath[_outputFilesBasePath.length() - 1] != '/') _outputFilesBasePath += '/';
      }

      // Data
      string         _outputFilesBasePath; // Base path for all destination (output) files
      vector<string> _outputFiles;         // List of output files
      vector<char>   _userData;            // User data
  };


  //!
  //! \brief A class used for loading archived data files
  //!        (without compression support)
  //!
  //! The size of the archive file must not exceed the capacity of uint32_t (4GB).
  //!
  //! \api experiment
  class VGE_EXPORT CArchiveDataSource : VCLASS_UNCOPYABLE {
    public:
      //!
      //! \brief Structure for storing informations about the files
      //!        in the archive
      //!
      //! \api experiment
      struct TFileListEntry {
        string   name;  //!< Name of the file (may has relative path in it)
        uint32_t start; //!< Start position of the file in the istream (in bytes)
        uint32_t size;  //!< Size of the file (in bytes)

        inline TFileListEntry(const string name_, uint32_t start_, uint32_t size_)
        : name(name_), start(start_), size(size_)
        {}
      };

      //! List of files in the archive
      typedef vector<TFileListEntry> CFileList;

    public:
      //!
      //! Construct a new CDeArchiver instance using the given parameters.
      //! \param[in] archiveFileName Archive file to be opened (the archive must
      //!                            be made using the CArchiver class). If the
      //!                            userInputStream parameter is not NULL, this
      //!                            parameter will be only used to name the stream
      //!                            (no new file will be opened, the given istream
      //!                            will be just used).
      //! \param[in] userInputStream If NULL, a new instance of std::ifstream will
      //!                            used to load the archive. If not NULL, the
      //!                            given istream instance will be used, in this
      //!                            case the given istream instance must be already
      //!                            initilized properly and points to the beginning
      //!                            of a valid archive data bytes that was made
      //!                            using the CArchiver class. The given istream
      //!                            must support seeking for this CArchiveDataSource
      //!                            class to work properly.
      CArchiveDataSource(const string &archiveFileName = "", istream *userInputStream = NULL);
      //!
      //! Standard dtor.
      ~CArchiveDataSource() throw();

      // Initialization functions
      //!
      //! (Re-)open a new archive.
      //! \param[in] archiveFileName Archive file to be opened (the archive must
      //!                            be made using the CArchiver class). If the
      //!                            userInputStream parameter is not NULL, this
      //!                            parameter will be only used to name the stream
      //!                            (no new file will be opened, the given istream
      //!                            will be just used).
      //! \param[in] userInputStream If NULL, a new instance of std::ifstream will
      //!                            used to load the archive. If not NULL, the
      //!                            given istream instance will be used, in this
      //!                            case the given istream instance must be already
      //!                            initilized properly and points to the beginning
      //!                            of a valid archive data bytes that was made
      //!                            using the CArchiver class. The given istream
      //!                            must support seeking for this CArchiveDataSource
      //!                            class to work properly.
      //!
      //! \warning Never use the given istream instance (userInputStream) from
      //!          other thread than the thread in which this CArchiveDataSource
      //!          instance is running.
      //! \warning This will make the previous CDataSource instance returned by
      //!          GetFile() become invalid.
      void Open(const string &archiveFileName, istream *userInputStream = NULL);
      //!
      //! Close the archive.
      //!
      //! \warning This will make the previous CDataSource instance returned by
      //!          GetFile() become invalid.
      void Close();

      // Data source functions
      //!
      //! Get some information about the files in the archive.
      inline const CFileList &GetFiles() const
      { return(_fileList); }
      //!
      //! Get the user data embedded in the archive.
      inline const vector<char> &GetUserData() const
      { return(_userData); }
      //!
      //! Get the user data embedded in the archive.
      const string GetUserDataStr() const;
      //!
      //! Returns a CDataSource instance to be used for loading the data.
      //!
      //! \warning Never use the returned CDataSource instance from more than
      //!          one thread.
      //! \warning This will make any previous CDataSource instance returned by
      //!          this function become invalid.
      CDataSource &GetFile(const string &fileName);

    protected:
      string          _archiveFN; // Archive file name
      bool            _ownIS;     // Flag to indicate if we own the _is instance below or not
      ios::iostate    _oex;       // Original exception flags of the _is instance below
      istream        *_is;        // The input stream to be used
      CDataSource    *_ds;        // The data source to be returned
      CFileList       _fileList;  // List of files
      vector<char>    _userData;  // User data embedded in the archive
  };

} // namespace vge
#endif
