// ############################################################################################
//  mcatedit-MCBuilder.h
//  Definition and implementation of mcatedit's CMCBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _UTIL___MCATEDIT_MCBUILDER_H_
#define _UTIL___MCATEDIT_MCBUILDER_H_

// VortexGE headers
#include <VMain.h>


//
// Used for searching in the catalog
//
struct TSearchOption {
  enum EFindIn       { OriginalMessage, TranslatedMessage };
  enum EMatchingRule { MatchOneWord, MatchAllWords, MatchExactSequence };

  EFindIn       findIn;
  EMatchingRule matchingRule;
  string        wordsToFind;
  bool          caseSensitive;
};

struct TSearchResult {
  uint32_t resultCurIndex;
  uint32_t resultTotIndex;

  string   fullText;
};


//
// Our message catalog builder class
//
class CMCBuilder {
  public:
    // Generic functions
    //
    // Check if the message catalog is valid (the catalog is open)
    inline bool IsOpen()
    { return(!path.empty()); }
    //
    // Get the current catalog path
    inline const string &CurrentPath() const
    { return(path); }
    //
    // Clear the message catalog (thus makes it empty)
    void Clear();

    // Searching functions
    //
    // Initiate search in the message catalog
    uint32_t Search_Start(const TSearchOption &opt);
    //
    // Select the first match
    bool Search_SelectFirstMatch();
    //
    // Select the previous match
    bool Search_SelectPrevMatch();
    //
    // Select the next match
    bool Search_SelectNextMatch();
    //
    // Select the last match
    bool Search_SelectLastMatch();
    //
    // Get the current selected search result
    bool Search_GetResult(TSearchResult &result);

    // Catalog I/O functions
    //
    // Create a new catalog to be saved in the given path
    void NewCatalog(const string &path_);
    //
    // Open a catalog from the given path
    void OpenCatalog(const string &path_);
    //
    // Save the catalog to the given path (if specified) or to the path set in the
    // previous call to newCatalog() or openCatalog()
    void SaveCatalog(const string &path_ = "");

    // Key/ID information functions
    //
    // Check if the key table is empty.
    inline bool HasNoKey() const
    { return(keys.msg.empty()); }
    //
    // Returns the total number of stored keys.
    inline uint32_t GetTotalStoredKeys() const
    { return(keys.msg.size()); }
    //
    // Returns the last key index.
    inline uint32_t GetLastKeyIndex() const
    { return(keys.msg.size() - 1); }
    //
    // Returns the index of the currently selected key.
    inline uint32_t GetCurrentKeyIndex() const
    { return(curKey); }
    //
    // Returns the ID of the currently selected key.
    inline uint32_t GetCurrentKeyID() const
    { return(keys.msg[curKey].id); }
    //
    // Returns the text of the currently selected key.
    inline const string &GetCurrentKeyText() const
    { return(keys.msg[curKey].text); }
    //
    // Check if the given key (using its ID) has ever been translated or not.
    bool KeyHasTranslations(uint32_t id) const;
    //
    // Check if the currently selected key has ever been translated or not.
    inline bool CurrentKeyHasTranslations() const
    { return(KeyHasTranslations(GetCurrentKeyID())); }

    // Key/ID management/manipulation functions
    //
    // Direct access to the key's optional information string.
    inline string &KeyInfo()
    { return(keys.info); }
    //
    // Direct access to the key's optional information string.
    inline const string &KeyInfo() const
    { return(keys.info); }
    // Direct access to the key's optional extra data.
    inline vector<char> &KeyExtraData()
    { return(keys.extra); }
    //
    // Direct access to the key's optional extra data.
    inline const vector<char> &KeyExtraData() const
    { return(keys.extra); }
    //
    // Creater a new empty key.
    void CreateNewKey();
    //
    // Delete currently selected key.
    void DeleteCurrentKey();
    //
    // Select the first key.
    bool SelectFirstKey();
    //
    // Select the previous key.
    bool SelectPreviousKey();
    //
    // Select the next key.
    bool SelectNextKey();
    //
    // Select the last key.
    bool SelectLastKey();
    //
    // Set the text of the currently selected key.
    inline void SetCurrentKeyText(const string &text)
    { keys.msg[curKey].text = text; }

    // Translated message information functions
    //
    // Check if the translation tables are empty.
    inline bool HasNoTranslation() const
    { return(translations.empty()); }
    //
    // Check if the catalog as a translation table for the given locale name.
    bool HasTranslationTable(const string &localeName) const;
    //
    // Check if the translation table for the given locale name is empty (or not exist).
    bool IsEmptyTranslationTable(const string &localeName) const;
    //
    // List all locale names for all translation tables into the given string vector.
    void ListAllLocales(vector<string> &dst);
    //
    // Get the locale name of the currently selected translation table.
    inline const string &GetCurrentTranslationTableLocale() const
    { return(translations[curTrans].locale); }
    //
    // Get the numeric ID of the selected translation text in the currently selected translation table.
    inline uint32_t GetCurrentTranslationTableID() const
    { return(curTransNI); }
    //
    // Access to the text of the selected translation text in the currently selected translation table.
    const string &GetCurrentTranslationTableText() const;

    // Translated message management/manipulation functions
    //
    // Select the translation table for the given locale name.
    bool SelectTranslationTable(const string &localeName);
    //
    // Direct access to the translation table for the given locale name.
    _TTranslations *GetTranslationTable(const string &localeName);
    //
    // Create a new translation table for the given locale name.
    inline void AddNewTranslationTable(const string &localeName)
    { translations.push_back(localeName); }
    //
    // Delete the translation table of the given locale name.
    void DeleteTranslationTable(const string &localeName);
    //
    // Select the first untranslated message text in the currently selected translation table.
    bool SelectFirstUntranslatedMessage();
    //
    // Select the previous untranslated message text in the currently selected translation table.
    bool SelectPreviousUntranslatedMessage();
    //
    // Select the next untranslated message text in the currently selected translation table.
    bool SelectNextUntranslatedMessage();
    //
    // Select the last untranslated message text in the currently selected translation table.
    bool SelectLastUntranslatedMessage();
    //
    // Search and select for untranslated message text in the currently selected translation table.
    bool SearchAndSelectUntranslatedMessage(size_t startIndex, size_t endIndex);
    //
    // Set the text of the selected translation text in the currently selected translation table.
    void SetCurrentTranslationText(const string &text);

  protected:
    // The catalog
    string                 path;         // Path to the directory in which the catalog will be saved
    _TKeys                 keys;         // Hold the keys
    vector<_TTranslations> translations; // Hold the translated messages for various locales
  
    // Used for editing the catalog
    uint32_t curKey;     // Index to the currently displayed keys
    uint32_t curTrans;   // Index to the currently translation array
    uint32_t curTransNI; // Numeric ID to the currently displayed translation

    // For searching
    struct _TSearchResults {
      vector<const string*> text;
      vector<uint32_t>      index;
      uint32_t              current;
    } _sresult;
};

#endif
