// bym.cc
// Generate C++ code for Bayer Matrix in two forms (namespace and struct)
//
// Copyright (C) 2004-2006 Aloysius Indrayanto
// This small utility is licensed under the GNU LGPL
//
// Usage: ./bym [> out-file]
// 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


// Uncomment this line to use 255 as the scaling factor (instead of 31)
//#define USE_8BIT_PRECISION


// This function was taken from :
//     Ordered Dithering
//     Stephen Hawley
//     Graphics Gems
//     Academic Press, 1990
//
// This function is used for calculating a value from a Bayer dithering matrice
// of which dimension are (2^order) by (2^order)
//
inline uint calcDitherValue(uint x, uint y, uint order) {
  uint d = 0;
  while (order-- > 0) {
    uint xlsb = x & 0x00000001;
    uint ylsb = y & 0x00000001;
    uint xyxr = xlsb ^ ylsb;
    uint dshf = d << 1;

    x >>= 1;
    y >>= 1;

    d = ((dshf | xyxr) << 1) | ylsb;
  }
  return(d);
}

// Calculate Bayer matrix utilizing the function above
inline void bayerMatrix(uint*& mat, uint& size)
{
  // Order 1 gives smoothest appereance but also gives minimum colors
  // Order 2 gives smoother appereance may not as colorfull as order 3
  // Order 3 is reccomended for the best colors
  // Order > 3 useless since it won't give noticeable visual improvements
  const uint order = 3;
  const uint msize = (1 << order);

  static uint matrix[msize * msize];
  static bool created = false;

  if(!created) {
    float m = -100000.0f;

    for(uint i = 0; i < (msize  * msize); i++) {
      matrix[i] = calcDitherValue(i / msize, i % msize, order);
      if(matrix[i] > m) m = matrix[i];
    }

    for(uint i = 0; i < (msize  * msize); i++)
#ifdef USE_8BIT_PRECISION
      matrix[i] = uint(roundf(matrix[i] / m * 255.0f));
#else
      matrix[i] = uint(roundf(matrix[i] / m * 31.0f));
#endif

    created = true;
  }

  mat  = matrix;
  size = msize;
}


// Main program
int main()
{
  uint* matrix;
  uint  size;
  bayerMatrix(matrix, size);

  // The namespace version
  printf("\nnamespace BayerMatrix {\n");
  printf("  static const uint8_t size       = %d;\n", size);
  printf("  static const uint8_t data[%d][%d] =\n", size, size);
  printf("    { { ");

  for(uint i = 0; i < size; i++) {
    for(uint j = 0; j < size; j++) {
#ifdef USE_8BIT_PRECISION
      if(j == (size - 1))
        printf("%3d ", matrix[i*size + j]);
      else
        printf("%3d, ", matrix[i*size + j]);
#else
      if(j == (size - 1))
        printf("%2d ", matrix[i*size + j]);
      else
        printf("%2d, ", matrix[i*size + j]);
#endif
    }
    if(i == (size - 1))
      printf("} };\n");
    else
      printf("}, \n      { ");
  }
  printf("};\n\n\n");

  // The struct version
  printf("struct TBayerMatrix {\n");
  printf("  static const uint8_t data[%d][%d];\n", size, size);
  printf("  static const uint8_t size = %d;\n", size);
  printf("};\n");

  printf("const uint8_t TBayerMatrix::data[%d][%d] =\n", size, size);
  printf("  { { ");

  for(uint i = 0; i < size; i++) {
    for(uint j = 0; j < size; j++) {
#ifdef USE_8BIT_PRECISION
      if(j == (size - 1))
        printf("%3d ", matrix[i*size + j]);
      else
        printf("%3d, ", matrix[i*size + j]);
#else
      if(j == (size - 1))
        printf("%2d ", matrix[i*size + j]);
      else
        printf("%2d, ", matrix[i*size + j]);
#endif
    }
    if(i == (size - 1))
      printf("} };\n");
    else
      printf("}, \n    { ");
  }
  printf("\n");

  return(0);
}
