// gcs.cc
// Read the file "CaseFolding.txt" in the same directory and
// generate C++ code to be used by VortexGE
//
// Copyright (C) 2004-2006 Aloysius Indrayanto
// This small utility is licensed under the GNU LGPL
//
// Usage: ./gcs [> out-file]
//
// The file "CaseFolding.txt" is downloaded from :
//     http://www.unicode.org/Public/UNIDATA/CaseFolding.txt
//
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
using namespace std;

#include <stdio.h>
#include <stdlib.h>


typedef uint32_t Utf32;

struct _TCFT {
  Utf32         src;
  vector<Utf32> dst;
};


static void _tokenize(vector<string> &tokens, const string &source, const string &delimiters)
{
  string::size_type lastPos = source.find_first_not_of(delimiters, 0);
  string::size_type pos     = source.find_first_of(delimiters, lastPos);
  tokens.clear();
  while((pos != string::npos) || (lastPos != string::npos)) {
    tokens.push_back(source.substr(lastPos, pos - lastPos));
    lastPos = source.find_first_not_of(delimiters, pos);
    pos     = source.find_first_of(delimiters, lastPos);
  }
}

static const string _trimSpace(const string &source)
{
  string::size_type sPos = source.find_first_not_of(" ", 0);
  string::size_type ePos = source.find_last_not_of(" ", source.size());
  return(source.substr(sPos, ePos - sPos + 1));
}

static void _printTable(const vector<_TCFT> &mapData, const string &varName, size_t maxCount)
{
  const vector<_TCFT> &_print = mapData;
  uint32_t             cCount = 0;

  printf("static const _TCaseFoldingMap %s[] = {\n", varName.c_str());
  for(size_t i = 0; i < _print.size(); ++i) {
    // Print the source
    printf("  {%5d,{", _print[i].src);
    // Print the destination
    const size_t dsz = _print[i].dst.size();
    for(size_t j = 0; j < dsz; ++j) {
      printf("%5d", _print[i].dst[j]);
      if(j != dsz-1) printf(",");
    }
    if(dsz < maxCount) {
      printf(",");
      for(size_t j = 0; j < maxCount-dsz; ++j) {
        printf("%5d", 0);
        if(j != maxCount-dsz-1) printf(",");
      }
    }
    // Print the destination count
    printf("},%d}", dsz);
    if(i == _print.size()-1)
      printf("\n};\n");
    else {
      cCount++;
      if(cCount < 2)
        printf(",");
      else {
        cCount = 0;
        printf(",\n");
      }
    }
  }
}


int main(int argc, char *argv[])
{
  // Case-folding tables
  vector<_TCFT> _C, _S, _F, _T;

  // Open the file
  ifstream ifs;
  ifs.open("gcs-CaseFolding.txt");

  // Process the file
  while(!ifs.eof()) {
    // Read a line
    string line;
    getline(ifs, line);
    if(line.empty() || line[0]=='#') continue;
    // Tokenize it
    vector<string> tokens;
    _tokenize(tokens, line, ";");
    if(tokens.size() < 3) continue;
    // Get the destination unicode
    vector<string> dst;
    _tokenize(dst, tokens[2], " ");
    if(!dst.size()) continue;
    // Get the source unicode and the case-folding mode
    const string &src = _trimSpace(tokens[0]);
    const string &cfm = _trimSpace(tokens[1]);
    // Convert them to case-folding table entry
    _TCFT cft;
    cft.src = strtol(src.c_str(), NULL, 16);
    for(size_t i = 0; i < dst.size(); ++i)
      cft.dst.push_back(strtol(dst[i].c_str(), NULL, 16));
    // Store to table
         if(cfm == "C") _C.push_back(cft);
    else if(cfm == "S") _S.push_back(cft);
    else if(cfm == "F") _F.push_back(cft);
    else if(cfm == "T") _T.push_back(cft);
  }

  // Find the maximum count
  size_t maxCount = 0;
  for(size_t i = 0; i < _C.size(); ++i) if(_C[i].dst.size() > maxCount) maxCount = _C[i].dst.size();
  for(size_t i = 0; i < _S.size(); ++i) if(_S[i].dst.size() > maxCount) maxCount = _S[i].dst.size();
  for(size_t i = 0; i < _F.size(); ++i) if(_F[i].dst.size() > maxCount) maxCount = _F[i].dst.size();
  for(size_t i = 0; i < _T.size(); ++i) if(_T[i].dst.size() > maxCount) maxCount = _T[i].dst.size();

  // Print the tables
  printf("\n//");
  printf("\n// A structure used for storing case-folding map entry");
  printf("\n//\n");
  printf("struct _TCaseFoldingMap {\n");
  printf("  CUString::Utf32 from;\n");
  printf("  CUString::Utf32 to[%d];\n", maxCount);
  printf("  uint32_t        toCount;\n");
  printf("};\n\n");
  printf("\n//");
  printf("\n// The data below was generated from a file which was downloaded from :");
  printf("\n//     http://www.unicode.org/Public/UNIDATA/CaseFolding.txt");
  printf("\n//\n");
  printf("\n// Common case-folding, common mappings shared by both simple and full mappings\n");
  _printTable(_C, "_ccf", maxCount);
  printf("\n// Full case-folding, mappings that cause strings to grow in length\n");
  _printTable(_F, "_fcf", maxCount);
  printf("\n// Simple case-folding, mappings to single characters (different from full case-folding)\n");
  _printTable(_S, "_scf", maxCount);
  printf("\n// Special case-folding, for example: uppercase I and dotted uppercase I in Turkic languages (tr & az locales)\n");
  _printTable(_T, "_tcf", maxCount);
  printf("\n");

  // Done
  return(0);
}
